setMethod("as.vector",
          signature(x = "externalVector"),
          function (x, mode)
      {
          if (missing(mode))
              mode <- "any"
          type <- class(as.vector(defaultElement(x), mode))
          if (type != class(defaultElement(x)))
              x <- asEach(x, type, arrayOnly=TRUE)
          ## strip attributes
          nonSlotAttributes(x) <- NULL
          x
      })

setMethod("as.logical",
          signature(x="externalVector"),
          function (x, ...)
          as.vector(x, "logical"))

setMethod("as.integer",
          signature(x="externalVector"),
          function (x, ...)
          as.vector(x, "integer"))

setMethod("as.numeric",
          signature(x="externalVector"),
          function (x, ...)
          as.vector(x, "numeric"))

setMethod("as.double",
          signature(x="externalVector"),
          function (x, ...)
          as.vector(x, "numeric"))

setMethod("as.single",
          signature(x="externalVector"),
          function (x, ...)
      {
          x <- as.vector(x, "numeric")
          attr(x, "Csingle") <- TRUE
          x
      })

setMethod("as.complex",
          signature(x="externalVector"),
          function (x, ...)
          as.vector(x, "complex"))

setMethod("as.character",
          signature(x="externalVector"),
          function (x) ## Note: does not have ... in generic
          as.vector(x, "character"))

setMethod("as.list",
          signature(x="externalVector"),
          function (x, ...)
          as.vector(x, "list"))

setMethod("as.matrix",
          signature(x = "externalVector"),
          function (x)
      {
          if (length(dim(x)) != 2) {
              nm <- names(x)
              dim(x) <- c(length(x), 1)
              if (!is.null(nm))
                  dimnames(x) <- list(nm, NULL)
          }
          ## strip attributes
          nonSlotAttributes(x) <- NULL
          x
      })

## coercion to basic R vector data types

setGeneric("as.Rvector",
               function(x)
               standardGeneric("as.Rvector"))

setMethod("as.Rvector",
          "logical",
          function (x) x)

setMethod("as.Rvector",
          "integer",
          function (x) x)

setMethod("as.Rvector",
          "numeric",
          function (x) x)

setMethod("as.Rvector",
          "complex",
          function (x) x)

setMethod("as.Rvector",
          "character",
          function (x) x)

setMethod("as.Rvector",
          "list",
          function (x) x)

setMethod("as.Rvector",
          "matrix",
          function (x) x)

setMethod("as.Rvector",
          "array",
          function (x) x)

setMethod("as.Rvector", "externalVector",
          function(x) as.Rvector(force(x)))

setMethod("as.Rvector",
          signature(x="externalVectorWithStorage"),
          function (x)
      {
          Dim <- dim(x)
          nDim <- length(Dim)
          Names <- names(x)
          saveAttribs <- nonSlotAttributes(x)
          if (nDim > 0) {
              DimNames <- dimnames(x)
              args <- vector('list', nDim+1)
              args[[1]] <- x
              args[-1] <- new("allIndex")
              x <- do.call('[', args)
              nonSlotAttributes(x) <- saveAttribs
              dim(x) <- Dim
              dimnames(x) <- DimNames
          } else {
              nonSlotAttributes(x) <- saveAttribs
              x <- x[new("allIndex")]
          }
          names(x) <- Names
          x
      })

# coercion into basic R matrix
setGeneric("as.Rmatrix",
               function(x)
               standardGeneric("as.Rmatrix"))

setMethod("as.Rmatrix",
          signature(x="matrix"),
          function (x) x)

setMethod("as.Rmatrix", "ANY",
          function (x) as.matrix(as.Rvector(x)))

setMethod("asEach",
          signature(x="externalVector", type = "character"),
          function(x, type, arrayOnly)
      {
          type <- class(as.vector(defaultElement(x), type))
          if (class(defaultElement(x)) == type)
              x
          else {
              x <- force(x)
              callNextmethod()
          }
      })

setMethod("as.data.frame",
          signature(x="externalVector"),
          function(x, row.names, optional)
          as.data.frame(as.Rmatrix(x), row.names, optional))

setMethod("c",
          signature(x="externalVector"),
          function(x, ..., recursive)
      {
          c(as.Rvector(x),
            c(..., recursive=recursive), recursive=recursive)
      })

setAs("externalVector", "vector",
      function(from, to) as.Rvector(from))

setAs("externalVector", "matrix",
      function(from, to) as.Rvector(as.matrix(from)))

setAs("vector", "externalVector",
      function(from, to)
  {
      to <- externalVector(class(from), length=length(from))
      to[] <- from
      d <- dim(to) <- dim(from)
      if (!is.null(d))
          dimnames(to) <- dimnames(from)
      nm <- names(from)
      if (!is.null(nm))
          names(to) <- nm
      to
  })

setAs("matrix", "externalVector",
      function(from, to)
  {
      to <- externalMatrix(from,
                           nrow = nrow(from),
                           ncol = ncol(from),
                           dimnames = dimnames(from))
      nm <- names(from)
      if (!is.null(nm))
          names(to) <- nm
      to
  })
