#include "storageMethods.h"

SEXP
getStorageExternalPtr(SEXP storage)
{
    SEXP ptr = GET_STORAGE_EXT_PTR_UNCHECKED(storage);
    void* addr = R_ExternalPtrAddr(ptr);
    if (addr == NULL) {
        storage = eval(PROTECT(lang2(install("reinitializePointer"),
                                             storage)),
                       R_GlobalEnv);
        UNPROTECT(1);
        ptr = GET_STORAGE_EXT_PTR_UNCHECKED(storage);
    }
    return ptr;
}

SEXP
getArithmeticAnswerStorage(SEXP xstorage,
                           const storageMethods_t* xmethods,
                           int nx,
                           SEXP ystorage,
                           const storageMethods_t* ymethods,
                           int ny, SEXPTYPE mode)
{
    if (nx >= ny) {
        return xmethods->alloc(xstorage,
                               allocVector(mode, 1), nx);
    } else {
        return ymethods->alloc(ystorage,
                               allocVector(mode, 1), ny);
    }
}

static primitiveElt_t*
defaultGetExtElt(const SEXPREC* storage, const storageMethods_t* methods, int i,
                 primitiveElt_t* val, int* warn)
{
    error("Element accessor not defined for mode %d", methods->mode);
    return NULL;
}

static const primitiveElt_t*
defaultSetExtElt(SEXP storage, const storageMethods_t* methods,
                 int i, const primitiveElt_t* val,
                 int* warn)
{
    error("Element value setting not allowed for mode %d", methods->mode);
    return NULL;
}

static primitiveElt_t*
defaultGetImaginaryFromAnyElt(const SEXPREC* storage,
                              const storageMethods_t* methods, int i,
                              primitiveElt_t* value, int* warn)
{
    return value;
}

static primitiveElt_t*
defaultGetImaginaryElt(const SEXPREC* storage,
                       const storageMethods_t* methods, int i,
                       primitiveElt_t* value, int* warn)
{
    methods->complexGetElt(storage, methods, i, value, warn);
    value->numeric = value->cmplx.i;
    return value;
}

static primitiveElt_t*
defaultGetLogicalFromIntegerElt(const SEXPREC* storage,
                                const storageMethods_t* methods, int i,
                                primitiveElt_t* value, int* warn)
{
    methods->integerGetElt(storage, methods, i, value, warn);
    value->logical = LogicalFromInteger(value->integer, warn);
    return value;
}

static primitiveElt_t*
defaultGetLogicalFromNumericElt(const SEXPREC* storage,
                                const storageMethods_t* methods, int i,
                                primitiveElt_t* value, int* warn)
{
    methods->numericGetElt(storage, methods, i, value, warn);
    value->logical = LogicalFromReal(value->numeric, warn);
    return value;
}

static primitiveElt_t*
defaultGetLogicalFromComplexElt(const SEXPREC* storage,
                                const storageMethods_t* methods, int i,
                                primitiveElt_t* value, int* warn)
{
    methods->complexGetElt(storage, methods, i, value, warn);
    value->logical = LogicalFromComplex(value->cmplx, warn);
    return value;
}

static primitiveElt_t*
defaultGetLogicalFromCharacterElt(const SEXPREC* storage,
                                  const storageMethods_t* methods, int i,
                                  primitiveElt_t* value, int* warn)
{
    methods->characterGetElt(storage, methods, i, value, warn);
    value->logical = LogicalFromString(value->character, warn);
    return value;
}

static primitiveElt_t*
defaultGetLogicalFromAnyElt(const SEXPREC* storage,
                            const storageMethods_t* methods, int i,
                            primitiveElt_t* value, int* warn)
{
    methods->anyGetElt(storage, methods, i, value, warn);
    value->logical = asLogical(value->any);
    return value;
}

/* static primitiveElt_t* */
/* defaultGetIntegerFromLogicalElt(const SEXPREC* storage, */
/*                                 const storageMethods_t* methods, int i, */
/*                                 primitiveElt_t* value, int* warn) */
/* { */
/*     methods->logicalGetElt(storage, methods, i, value, warn); */
/*     return value; */
/* } */

static primitiveElt_t*
defaultGetIntegerFromNumericElt(const SEXPREC* storage,
                                const storageMethods_t* methods, int i,
                                primitiveElt_t* value, int* warn)
{
    methods->numericGetElt(storage, methods, i, value, warn);
    value->integer = IntegerFromReal(value->numeric, warn);
    return value;
}

static primitiveElt_t*
defaultGetIntegerFromComplexElt(const SEXPREC* storage,
                                const storageMethods_t* methods, int i,
                                primitiveElt_t* value, int* warn)
{
    methods->complexGetElt(storage, methods, i, value, warn);
    value->integer = IntegerFromComplex(value->cmplx, warn);
    return value;
}

static primitiveElt_t*
defaultGetIntegerFromCharacterElt(const SEXPREC* storage,
                                  const storageMethods_t* methods, int i,
                                  primitiveElt_t* value, int* warn)
{
    methods->characterGetElt(storage, methods, i, value, warn);
    value->integer = IntegerFromString(value->character, warn);
    return value;
}

static primitiveElt_t*
defaultGetIntegerFromAnyElt(const SEXPREC* storage,
                            const storageMethods_t* methods, int i,
                            primitiveElt_t* value, int* warn)
{
    methods->anyGetElt(storage, methods, i, value, warn);
    value->integer = asInteger(value->any);
    return value;
}

static primitiveElt_t*
defaultGetNumericFromLogicalElt(const SEXPREC* storage,
                                const storageMethods_t* methods, int i,
                                primitiveElt_t* value, int* warn)
{
    methods->logicalGetElt(storage, methods, i, value, warn);
    value->numeric = RealFromLogical(value->logical, warn);
    return value;
}

static primitiveElt_t*
defaultGetNumericFromIntegerElt(const SEXPREC* storage,
                                const storageMethods_t* methods, int i,
                                primitiveElt_t* value, int* warn)
{
    methods->integerGetElt(storage, methods, i, value, warn);
    value->numeric = RealFromInteger(value->integer, warn);
    return value;
}

static primitiveElt_t*
defaultGetNumericFromComplexElt(const SEXPREC* storage,
                                const storageMethods_t* methods, int i,
                                primitiveElt_t* value, int* warn)
{
    methods->complexGetElt(storage, methods, i, value, warn);
    value->numeric = RealFromComplex(value->cmplx, warn);
    return value;
}

static primitiveElt_t*
defaultGetNumericFromCharacterElt(const SEXPREC* storage,
                                  const storageMethods_t* methods, int i,
                                  primitiveElt_t* value, int* warn)
{
    methods->characterGetElt(storage, methods, i, value, warn);
    value->numeric = RealFromString(value->character, warn);
    return value;
}

static primitiveElt_t*
defaultGetNumericFromAnyElt(const SEXPREC* storage,
                            const storageMethods_t* methods, int i,
                            primitiveElt_t* value, int* warn)
{
    methods->anyGetElt(storage, methods, i, value, warn);
    value->numeric = asReal(value->any);
    return value;
}

static primitiveElt_t*
defaultGetComplexFromLogicalElt(const SEXPREC* storage,
                                const storageMethods_t* methods, int i,
                                primitiveElt_t* value, int* warn)
{
    methods->logicalGetElt(storage, methods, i, value, warn);
    value->cmplx = ComplexFromLogical(value->logical, warn);
    return value;
}

static primitiveElt_t*
defaultGetComplexFromIntegerElt(const SEXPREC* storage,
                                const storageMethods_t* methods, int i,
                                primitiveElt_t* value, int* warn)
{
    methods->integerGetElt(storage, methods, i, value, warn);
    value->cmplx = ComplexFromInteger(value->integer, warn);
    return value;
}

static primitiveElt_t*
defaultGetComplexFromNumericElt(const SEXPREC* storage,
                                const storageMethods_t* methods, int i,
                                primitiveElt_t* value, int* warn)
{
    methods->numericGetElt(storage, methods, i, value, warn);
    value->cmplx = ComplexFromReal(value->numeric, warn);
    return value;
}

static primitiveElt_t*
defaultGetComplexFromCharacterElt(const SEXPREC* storage,
                                  const storageMethods_t* methods, int i,
                                  primitiveElt_t* value, int* warn)
{
    methods->characterGetElt(storage, methods, i, value, warn);
    value->cmplx = ComplexFromString(value->character, warn);
    return value;
}

static primitiveElt_t*
defaultGetComplexFromAnyElt(const SEXPREC* storage,
                            const storageMethods_t* methods, int i,
                            primitiveElt_t* value, int* warn)
{
    methods->anyGetElt(storage, methods, i, value, warn);
    value->cmplx = asComplex(value->any);
    return value;
}

static primitiveElt_t*
defaultGetCharacterFromLogicalElt(const SEXPREC* storage,
                                  const storageMethods_t* methods, int i,
                                  primitiveElt_t* value, int* warn)
{
    methods->logicalGetElt(storage, methods, i, value, warn);
    value->character = StringFromLogical(value->logical, warn);
    return value;
}

static primitiveElt_t*
defaultGetCharacterFromIntegerElt(const SEXPREC* storage,
                                  const storageMethods_t* methods, int i,
                                  primitiveElt_t* value, int* warn)
{
    methods->integerGetElt(storage, methods, i, value, warn);
    value->character = StringFromInteger(value->integer, warn);
    return value;
}

static primitiveElt_t*
defaultGetCharacterFromNumericElt(const SEXPREC* storage,
                                  const storageMethods_t* methods, int i,
                                  primitiveElt_t* value, int* warn)
{
    methods->numericGetElt(storage, methods, i, value, warn);
    value->character = StringFromReal(value->numeric, warn);
    return value;
}

static primitiveElt_t*
defaultGetCharacterFromComplexElt(const SEXPREC* storage,
                                  const storageMethods_t* methods, int i,
                                  primitiveElt_t* value, int* warn)
{
    methods->complexGetElt(storage, methods, i, value, warn);
    value->character = StringFromComplex(value->cmplx, warn);
    return value;
}

static primitiveElt_t*
defaultGetCharacterFromAnyElt(const SEXPREC* storage,
                            const storageMethods_t* methods, int i,
                            primitiveElt_t* value, int* warn)
{
    methods->anyGetElt(storage, methods, i, value, warn);
    value->character = asChar(value->any);
    return value;
}

static primitiveElt_t*
defaultGetAnyFromLogicalElt(const SEXPREC* storage,
                            const storageMethods_t* methods, int i,
                            primitiveElt_t* value, int* warn)
{
    methods->logicalGetElt(storage, methods, i, value, warn);
    value->any = ScalarLogical(value->logical);
    return value;
}

static primitiveElt_t*
defaultGetAnyFromIntegerElt(const SEXPREC* storage,
                            const storageMethods_t* methods, int i,
                            primitiveElt_t* value, int* warn)
{
    methods->integerGetElt(storage, methods, i, value, warn);
    value->any = ScalarInteger(value->integer);
    return value;
}

static primitiveElt_t*
defaultGetAnyFromNumericElt(const SEXPREC* storage,
                            const storageMethods_t* methods, int i,
                            primitiveElt_t* value, int* warn)
{
    methods->numericGetElt(storage, methods, i, value, warn);
    value->any = ScalarReal(value->numeric);
    return value;
}

static primitiveElt_t*
defaultGetAnyFromComplexElt(const SEXPREC* storage,
                            const storageMethods_t* methods, int i,
                            primitiveElt_t* value, int* warn)
{
    methods->complexGetElt(storage, methods, i, value, warn);
    value->any = ScalarComplex(value->cmplx);
    return value;
}

static primitiveElt_t*
defaultGetAnyFromCharacterElt(const SEXPREC* storage,
                              const storageMethods_t* methods, int i,
                              primitiveElt_t* value, int* warn)
{
    methods->characterGetElt(storage, methods, i, value, warn);
    value->any = ScalarString(value->character);
    return value;
}

/* static const primitiveElt_t* */
/* defaultSetLogicalToIntegerElt(SEXP storage, */
/*                               const storageMethods_t* methods, int i, */
/*                               const primitiveElt_t* value, int* warn) */
/* { */
/*     methods->integerSetElt(storage, methods, i, value, warn); */
/*     return value; */
/* } */

static const primitiveElt_t*
defaultSetAnyToImaginaryElt(SEXP storage,
                            const storageMethods_t* methods, int i,
                            const primitiveElt_t* value, int* warn)
{
    /* maybe set warn here? */
    return value;
}

static const primitiveElt_t*
defaultSetImaginaryElt(SEXP storage,
                       const storageMethods_t* methods, int i,
                       const primitiveElt_t* value, int* warn)
{
    primitiveElt_t value1;
    methods->complexGetElt(storage, methods, i, &value1, warn);
    value1.cmplx.i = value->numeric;
    methods->complexSetElt(storage, methods, i, &value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetLogicalToNumericElt(SEXP storage,
                              const storageMethods_t* methods, int i,
                              const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->numeric = RealFromLogical(value->logical, warn);
    methods->numericSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetLogicalToComplexElt(SEXP storage,
                              const storageMethods_t* methods, int i,
                              const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->cmplx = ComplexFromLogical(value->logical, warn);
    methods->complexSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetLogicalToCharacterElt(SEXP storage,
                                const storageMethods_t* methods, int i,
                                const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->character = StringFromLogical(value->logical, warn);
    methods->characterSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetLogicalToAnyElt(SEXP storage,
                          const storageMethods_t* methods, int i,
                          const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->any = PROTECT(ScalarLogical(value->logical));
    methods->anySetElt(storage, methods, i, value1, warn);
    UNPROTECT(1);
    return value;
}

static const primitiveElt_t*
defaultSetIntegerToLogicalElt(SEXP storage,
                              const storageMethods_t* methods, int i,
                              const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->logical = LogicalFromInteger(value->integer, warn);
    methods->logicalSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetIntegerToNumericElt(SEXP storage,
                              const storageMethods_t* methods, int i,
                              const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->numeric = RealFromInteger(value->integer, warn);
    methods->numericSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetIntegerToComplexElt(SEXP storage,
                              const storageMethods_t* methods, int i,
                              const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->cmplx = ComplexFromInteger(value->integer, warn);
    methods->complexSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetIntegerToCharacterElt(SEXP storage,
                                const storageMethods_t* methods, int i,
                                const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->character = StringFromInteger(value->integer, warn);
    methods->characterSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetIntegerToAnyElt(SEXP storage,
                          const storageMethods_t* methods, int i,
                          const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->any = PROTECT(ScalarInteger(value->integer));
    methods->anySetElt(storage, methods, i, value1, warn);
    UNPROTECT(1);
    return value;
}

static const primitiveElt_t*
defaultSetNumericToLogicalElt(SEXP storage,
                              const storageMethods_t* methods, int i,
                              const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->logical = LogicalFromReal(value->numeric, warn);
    methods->logicalSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetNumericToIntegerElt(SEXP storage,
                              const storageMethods_t* methods, int i,
                              const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->integer = IntegerFromReal(value->numeric, warn);
    methods->integerSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetNumericToComplexElt(SEXP storage,
                              const storageMethods_t* methods, int i,
                              const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->cmplx = ComplexFromReal(value->numeric, warn);
    methods->complexSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetNumericToCharacterElt(SEXP storage,
                                const storageMethods_t* methods, int i,
                                const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->character = StringFromReal(value->numeric, warn);
    methods->characterSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetNumericToAnyElt(SEXP storage,
                          const storageMethods_t* methods, int i,
                          const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->any = PROTECT(ScalarReal(value->numeric));
    methods->anySetElt(storage, methods, i, value1, warn);
    UNPROTECT(1);
    return value;
}

static const primitiveElt_t*
defaultSetComplexToLogicalElt(SEXP storage,
                              const storageMethods_t* methods, int i,
                              const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->logical = LogicalFromComplex(value->cmplx, warn);
    methods->logicalSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetComplexToIntegerElt(SEXP storage,
                              const storageMethods_t* methods, int i,
                              const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->integer = IntegerFromComplex(value->cmplx, warn);
    methods->integerSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetComplexToNumericElt(SEXP storage,
                              const storageMethods_t* methods, int i,
                              const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->numeric = RealFromComplex(value->cmplx, warn);
    methods->numericSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetComplexToCharacterElt(SEXP storage,
                                const storageMethods_t* methods, int i,
                                const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->character = StringFromComplex(value->cmplx, warn);
    methods->characterSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetComplexToAnyElt(SEXP storage,
                          const storageMethods_t* methods, int i,
                          const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->any = PROTECT(ScalarComplex(value->cmplx));
    methods->anySetElt(storage, methods, i, value1, warn);
    UNPROTECT(1);
    return value;
}

static const primitiveElt_t*
defaultSetCharacterToLogicalElt(SEXP storage,
                                const storageMethods_t* methods, int i,
                                const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->logical = LogicalFromString(value->character, warn);
    methods->logicalSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetCharacterToIntegerElt(SEXP storage,
                                const storageMethods_t* methods, int i,
                                const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->integer = IntegerFromString(value->character, warn);
    methods->integerSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetCharacterToNumericElt(SEXP storage,
                                const storageMethods_t* methods, int i,
                                const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->numeric = RealFromString(value->character, warn);
    methods->numericSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetCharacterToComplexElt(SEXP storage,
                                const storageMethods_t* methods, int i,
                                const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->cmplx = ComplexFromString(value->character, warn);
    methods->complexSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetCharacterToAnyElt(SEXP storage,
                          const storageMethods_t* methods, int i,
                          const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->any = PROTECT(ScalarString(value->character));
    methods->anySetElt(storage, methods, i, value1, warn);
    UNPROTECT(1);
    return value;
}

static const primitiveElt_t*
defaultSetAnyToLogicalElt(SEXP storage,
                          const storageMethods_t* methods, int i,
                          const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->logical = asLogical(value->any);
    methods->logicalSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetAnyToIntegerElt(SEXP storage,
                          const storageMethods_t* methods, int i,
                          const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->integer = asInteger(value->any);
    methods->integerSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetAnyToNumericElt(SEXP storage,
                          const storageMethods_t* methods, int i,
                          const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->numeric = asReal(value->any);
    methods->numericSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetAnyToComplexElt(SEXP storage,
                          const storageMethods_t* methods, int i,
                          const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->cmplx = asComplex(value->any);
    methods->complexSetElt(storage, methods, i, value1, warn);
    return value;
}

static const primitiveElt_t*
defaultSetAnyToCharacterElt(SEXP storage,
                          const storageMethods_t* methods, int i,
                          const primitiveElt_t* value, int* warn)
{
    primitiveElt_t tmp, *value1 = &tmp;
    value1->character = PROTECT(asChar(value->any));
    methods->characterSetElt(storage, methods, i, value1, warn);
    UNPROTECT(1);
    return value;
}

static primitiveElt_t*
defaultCommonGetMatrixElt(const SEXPREC* storage,
                          const storageMethods_t* methods, int i, int j,
                          int m, int n,
                          primitiveElt_t* value,
                          getExtElt_t getElt, int* warn)
{
    i += j*m;
    return getElt(storage, methods, i, value, warn);
}

static primitiveElt_t*
defaultGetExtMatrixElt(const SEXPREC* storage,
                       const storageMethods_t* methods, int i, int j,
                       int m, int n,
                       primitiveElt_t* value, int* warn)
{
    getExtElt_t getElt = methods->getExtElt;
    return defaultCommonGetMatrixElt(storage, methods, i, j, m, n, value,
                                     getElt, warn);
}

static primitiveElt_t*
defaultLogicalGetMatrixElt(const SEXPREC* storage,
                           const storageMethods_t* methods, int i, int j,
                           int m, int n,
                           primitiveElt_t* value, int* warn)
{
    getExtElt_t getElt = methods->logicalGetElt;
    return defaultCommonGetMatrixElt(storage, methods, i, j, m, n, value,
                                     getElt, warn);
}

static primitiveElt_t*
defaultIntegerGetMatrixElt(const SEXPREC* storage,
                           const storageMethods_t* methods, int i, int j,
                           int m, int n,
                           primitiveElt_t* value, int* warn)
{
    getExtElt_t getElt = methods->integerGetElt;
    return defaultCommonGetMatrixElt(storage, methods, i, j, m, n, value,
                                     getElt, warn);
}

static primitiveElt_t*
defaultNumericGetMatrixElt(const SEXPREC* storage,
                           const storageMethods_t* methods, int i, int j,
                           int m, int n,
                           primitiveElt_t* value, int* warn)
{
    getExtElt_t getElt = methods->numericGetElt;
    return defaultCommonGetMatrixElt(storage, methods, i, j, m, n, value,
                                     getElt, warn);
}

static primitiveElt_t*
defaultComplexGetMatrixElt(const SEXPREC* storage,
                           const storageMethods_t* methods, int i, int j,
                           int m, int n,
                           primitiveElt_t* value, int* warn)
{
    getExtElt_t getElt = methods->complexGetElt;
    return defaultCommonGetMatrixElt(storage, methods, i, j, m, n, value,
                                     getElt, warn);
}

static primitiveElt_t*
defaultImaginaryGetMatrixElt(const SEXPREC* storage,
                             const storageMethods_t* methods, int i, int j,
                             int m, int n,
                             primitiveElt_t* value, int* warn)
{
    getExtElt_t getElt = methods->imaginaryGetElt;
    return defaultCommonGetMatrixElt(storage, methods, i, j, m, n, value,
                                     getElt, warn);
}

static primitiveElt_t*
defaultCharacterGetMatrixElt(const SEXPREC* storage,
                             const storageMethods_t* methods, int i, int j,
                             int m, int n,
                             primitiveElt_t* value, int* warn)
{
    getExtElt_t getElt = methods->characterGetElt;
    return defaultCommonGetMatrixElt(storage, methods, i, j, m, n, value,
                                     getElt, warn);
}

static primitiveElt_t*
defaultAnyGetMatrixElt(const SEXPREC* storage,
                       const storageMethods_t* methods, int i, int j,
                       int m, int n,
                       primitiveElt_t* value, int* warn)
{
    getExtElt_t getElt = methods->anyGetElt;
    return defaultCommonGetMatrixElt(storage, methods, i, j, m, n, value,
                                     getElt, warn);
}

static const primitiveElt_t*
defaultCommonSetMatrixElt(SEXP storage,
                          const storageMethods_t* methods, int i, int j,
                          int m, int n,
                          const primitiveElt_t* value,
                          setExtElt_t setElt, int* warn)
{
    i += j*m;
    return setElt(storage, methods, i, value, warn);
}

static const primitiveElt_t*
defaultSetExtMatrixElt(SEXP storage,
                       const storageMethods_t* methods, int i, int j,
                       int m, int n,
                       const primitiveElt_t* value, int* warn)
{
    setExtElt_t setElt = methods->setExtElt;
    return defaultCommonSetMatrixElt(storage, methods, i, j, m, n, value,
                                     setElt, warn);
}

static const primitiveElt_t*
defaultLogicalSetMatrixElt(SEXP storage,
                           const storageMethods_t* methods, int i, int j,
                           int m, int n,
                           const primitiveElt_t* value, int* warn)
{
    setExtElt_t setElt = methods->logicalSetElt;
    return defaultCommonSetMatrixElt(storage, methods, i, j, m, n, value,
                                     setElt, warn);
}

static const primitiveElt_t*
defaultIntegerSetMatrixElt(SEXP storage,
                           const storageMethods_t* methods, int i, int j,
                           int m, int n,
                           const primitiveElt_t* value, int* warn)
{
    setExtElt_t setElt = methods->integerSetElt;
    return defaultCommonSetMatrixElt(storage, methods, i, j, m, n, value,
                                     setElt, warn);
}

static const primitiveElt_t*
defaultNumericSetMatrixElt(SEXP storage,
                           const storageMethods_t* methods, int i, int j,
                           int m, int n,
                           const primitiveElt_t* value, int* warn)
{
    setExtElt_t setElt = methods->numericSetElt;
    return defaultCommonSetMatrixElt(storage, methods, i, j, m, n, value,
                                     setElt, warn);
}

static const primitiveElt_t*
defaultComplexSetMatrixElt(SEXP storage,
                           const storageMethods_t* methods, int i, int j,
                           int m, int n,
                           const primitiveElt_t* value, int* warn)
{
    setExtElt_t setElt = methods->complexSetElt;
    return defaultCommonSetMatrixElt(storage, methods, i, j, m, n, value,
                                     setElt, warn);
}

static const primitiveElt_t*
defaultImaginarySetMatrixElt(SEXP storage,
                             const storageMethods_t* methods, int i, int j,
                             int m, int n,
                             const primitiveElt_t* value, int* warn)
{
    setExtElt_t setElt = methods->imaginarySetElt;
    return defaultCommonSetMatrixElt(storage, methods, i, j, m, n, value,
                                     setElt, warn);
}

static const primitiveElt_t*
defaultCharacterSetMatrixElt(SEXP storage,
                             const storageMethods_t* methods, int i, int j,
                             int m, int n,
                             const primitiveElt_t* value, int* warn)
{
    setExtElt_t setElt = methods->characterSetElt;
    return defaultCommonSetMatrixElt(storage, methods, i, j, m, n, value,
                                     setElt, warn);
}

static const primitiveElt_t*
defaultAnySetMatrixElt(SEXP storage,
                       const storageMethods_t* methods, int i, int j,
                       int m, int n,
                       const primitiveElt_t* value, int* warn)
{
    setExtElt_t setElt = methods->anySetElt;
    return defaultCommonSetMatrixElt(storage, methods, i, j, m, n, value,
                                     setElt, warn);
}

static SEXP
defaultCommonVectorSubset(const SEXPREC* storage, storageMethods_t* methods,
                          const SEXPREC* s, int nx, SEXP value, SEXPTYPE mode,
                          setRElt_t setRElt, getExtElt_t getExtElt,
                          const primitiveElt_t* NA_TYPE, int* warn)
{
    const SEXPREC* subtype = getIndexType(s);
    if (subtype == NAIndex) {
        int i;
        int ns = asInteger(GET_SLOT((SEXP)s, install("length")));
        if (value == R_NilValue)
            value = allocVector(mode, ns);
        PROTECT(value);
        for (i = 0; i < ns; i++)
            setRElt(value, i, NA_TYPE);
    } else if (subtype == allIndex) {
        primitiveElt_t elt;
        int i;
        if (value == R_NilValue)
            value = allocVector(mode, nx);
        PROTECT(value);
        for (i = 0; i < nx; i++) {
            getExtElt(storage, methods, i, &elt, warn);
            setRElt(value, i, &elt);
        }
    } else if (subtype == sliceIndex) {
        double* content = REAL(GET_SLOT((SEXP) s, install("content")));
        int start = (int) content[0];
        int ns = (int) content[1];
        int stride = (int) content[2];
        primitiveElt_t elt;
        int i;

        if (value == R_NilValue)
            value = allocVector(mode, ns);
        PROTECT(value);

        if (ns < nx)
            nx = ns;
        for (i = 0; i < nx; i++, start += stride) {
            getExtElt(storage, methods, start, &elt, warn);
            setRElt(value, i, &elt);
        }
        for (; i < ns; i++)
            setRElt(value, i, NA_TYPE);
    } else {
        int ns = length((SEXP) s);
        primitiveElt_t elt;
        int i = 0;
        int* indx = INTEGER((SEXP) s);

        if (value == R_NilValue)
            value = allocVector(mode, ns);
        PROTECT(value);
        for (i = 0; i < ns; i++) {
            int j = indx[i];
            if (j >= nx || j == NA_INTEGER)
                setRElt(value, i, NA_TYPE);
            else {
                getExtElt(storage, methods, j, &elt, warn);
                setRElt(value, i, &elt);
            }
        }
    }
    UNPROTECT(1);
    return value;
}

static SEXP
defaultVectorSubset(const SEXPREC* storage, storageMethods_t* methods,
                    const SEXPREC* s, int nx, SEXP value,
                    int* warn)
{
    setRElt_t setRElt = methods->setRElt;
    getExtElt_t getExtElt = methods->getExtElt;
    primitiveElt_t* NA_TYPE = &methods->NA;
    return defaultCommonVectorSubset(storage, methods, s, nx, value,
                                     methods->mode,
                                     setRElt, getExtElt, NA_TYPE, warn);
}

static SEXP
defaultNumericVectorSubset(const SEXPREC* storage, storageMethods_t* methods,
                           const SEXPREC* s, int nx, SEXP value,
                           int* warn)
{
    setRElt_t setRElt = methods->numericSetRElt;
    getExtElt_t getExtElt = methods->numericGetElt;
    primitiveElt_t NA;
    NA.numeric = NA_REAL;
    return defaultCommonVectorSubset(storage, methods, s, nx, value,
                                     REALSXP,
                                     setRElt, getExtElt, &NA, warn);
}

static SEXP
defaultComplexVectorSubset(const SEXPREC* storage, storageMethods_t* methods,
                           const SEXPREC* s, int nx, SEXP value,
                           int* warn)
{
    setRElt_t setRElt = methods->complexSetRElt;
    getExtElt_t getExtElt = methods->complexGetElt;
    primitiveElt_t NA;
    NA.cmplx.r = NA.cmplx.i = NA_REAL;
    return defaultCommonVectorSubset(storage, methods, s, nx, value,
                                     CPLXSXP,
                                     setRElt, getExtElt, &NA, warn);
}

static SEXP
defaultCharacterVectorSubset(const SEXPREC* storage, storageMethods_t* methods,
                           const SEXPREC* s, int nx, SEXP value,
                           int* warn)
{
    setRElt_t setRElt = methods->characterSetRElt;
    getExtElt_t getExtElt = methods->characterGetElt;
    primitiveElt_t NA;
    NA.character = NA_STRING;
    return defaultCommonVectorSubset(storage, methods, s, nx, value,
                                     STRSXP,
                                     setRElt, getExtElt, &NA, warn);
}

static SEXP
defaultLogicalVectorSubset(const SEXPREC* storage, storageMethods_t* methods,
                           const SEXPREC* s, int nx, SEXP value,
                           int* warn)
{
    setRElt_t setRElt = methods->logicalSetRElt;
    getExtElt_t getExtElt = methods->logicalGetElt;
    primitiveElt_t NA;
    NA.logical = NA_LOGICAL;
    return defaultCommonVectorSubset(storage, methods, s, nx, value,
                                     LGLSXP,
                                     setRElt, getExtElt, &NA, warn);
}

static SEXP
defaultIntegerVectorSubset(const SEXPREC* storage, storageMethods_t* methods,
                           const SEXPREC* s, int nx, SEXP value,
                           int* warn)
{
    setRElt_t setRElt = methods->integerSetRElt;
    getExtElt_t getExtElt = methods->integerGetElt;
    primitiveElt_t NA;
    NA.integer = NA_INTEGER;
    return defaultCommonVectorSubset(storage, methods, s, nx, value,
                                     INTSXP,
                                     setRElt, getExtElt, &NA, warn);
}

static SEXP
defaultAnyVectorSubset(const SEXPREC* storage, storageMethods_t* methods,
                       const SEXPREC* s, int nx, SEXP value,
                       int* warn)
{
    setRElt_t setRElt = methods->listSetRElt;
    getExtElt_t getExtElt = methods->anyGetElt;
    primitiveElt_t NA;
    NA.any = R_NilValue;
    return defaultCommonVectorSubset(storage, methods, s, nx, value,
                                     VECSXP,
                                     setRElt, getExtElt, &NA, warn);
}

#define SMALL_VALUE_LENGTH (5)

static const SEXPREC*
defaultCommonVectorSubassign(SEXP storage, storageMethods_t* methods,
                             const SEXPREC* s, int nx, const SEXPREC* value,
                             getRElt_t getRElt, setExtElt_t setExtElt,
                             int* warn)
{
    int nvalue = LENGTH((SEXP)value);
    const SEXPREC* subtype = getIndexType(s);
    if (subtype == noneIndex || subtype == NAIndex) {
        return value;
    } else if (subtype == allIndex) {
        primitiveElt_t elt;

        if (nvalue != nx &&
            nvalue <= SMALL_VALUE_LENGTH) {/* Do this only if nvalue is small */
            int i1;
            for (i1 = 0; i1 < nvalue; i1++) {
                int i;
                getRElt(value, i1, &elt);
                for (i = i1; i < nx; i += nvalue) {
                    setExtElt(storage, methods, i, &elt, warn);
                }
            }
        } else {
            int i, i1;
            for (i = 0, i1 = 0; i < nx; i++, i1++) {
                if (i1 == nvalue)
                    i1 = 0;
                getRElt(value, i1, &elt);
                setExtElt(storage, methods, i, &elt, warn);
            }
        }
    } else if (subtype == sliceIndex) {
        double* content = REAL(GET_SLOT((SEXP) s, install("content")));
        int start = (int) content[0];
        int ns = (int) content[1];
        int stride = (int) content[2];
        int end = start + ns*stride;
        primitiveElt_t elt;

        if (nvalue != ns &&
            nvalue <= SMALL_VALUE_LENGTH) {/* Do this only if nvalue is small */
            int i1;
            for (i1 = 0; i1 < nvalue; i1++) {
                int i, start1;
                int stride1 = stride*nvalue;
                getRElt(value, i1, &elt);
                for (i = i1, start1 = start+i1*stride;
                     i < end; i += nvalue, start1 += stride1) {
                    setExtElt(storage, methods, start1, &elt, warn);
                }
            }
        } else {
            int i, i1;
            for (i = 0, i1 = 0; i < ns; i++, i1++, start += stride) {
                if (i1 == nvalue)
                    i1 = 0;
                getRElt(value, i1, &elt);
                setExtElt(storage, methods, start, &elt, warn);
            }
        }
    } else {
        int ns = LENGTH((SEXP) s);
        int* indx = INTEGER((SEXP) s);
        primitiveElt_t elt;
        if (nvalue != ns &&
            nvalue <= SMALL_VALUE_LENGTH) {/* Do this only if nvalue is small */
            int i1;
            for (i1 = 0; i1 < nvalue; i1++) {
                int i;
                getRElt(value, i1, &elt);
                for (i = i1; i < ns; i += nvalue) {
                    setExtElt(storage, methods, indx[i], &elt, warn);
                }
            }
        } else {
            int i, i1;
            for (i = 0, i1 = 0; i < ns; i++, i1++) {
                if (i1 == nvalue)
                    i1 = 0;
                getRElt(value, i1, &elt);
                setExtElt(storage, methods, indx[i], &elt, warn);
            }
        }
    }
    return value;
}

static const SEXPREC*
defaultVectorSubassign(SEXP storage, storageMethods_t* methods,
                       const SEXPREC* s, int nx, const SEXPREC* value,
                       int* warn)
{
    getRElt_t getRElt = methods->getRElt;
    setExtElt_t setExtElt = methods->setExtElt;
    return defaultCommonVectorSubassign(storage, methods, s, nx, value,
                                        getRElt, setExtElt, warn);
}

static const SEXPREC*
defaultNumericVectorSubassign(SEXP storage, storageMethods_t* methods,
                              const SEXPREC* s, int nx, const SEXPREC* value,
                              int* warn)
{
    getRElt_t getRElt = methods->numericGetRElt;
    setExtElt_t setExtElt = methods->numericSetElt;
    return defaultCommonVectorSubassign(storage, methods, s, nx, value,
                                        getRElt, setExtElt, warn);
}

static const SEXPREC*
defaultIntegerVectorSubassign(SEXP storage, storageMethods_t* methods,
                              const SEXPREC* s, int nx, const SEXPREC* value,
                              int* warn)
{
    getRElt_t getRElt = methods->integerGetRElt;
    setExtElt_t setExtElt = methods->integerSetElt;
    return defaultCommonVectorSubassign(storage, methods, s, nx, value,
                                        getRElt, setExtElt, warn);
}

static const SEXPREC*
defaultComplexVectorSubassign(SEXP storage, storageMethods_t* methods,
                              const SEXPREC* s, int nx, const SEXPREC* value,
                              int* warn)
{
    getRElt_t getRElt = methods->complexGetRElt;
    setExtElt_t setExtElt = methods->complexSetElt;
    return defaultCommonVectorSubassign(storage, methods, s, nx, value,
                                        getRElt, setExtElt, warn);
}

static const SEXPREC*
defaultCharacterVectorSubassign(SEXP storage, storageMethods_t* methods,
                              const SEXPREC* s, int nx, const SEXPREC* value,
                              int* warn)
{
    getRElt_t getRElt = methods->characterGetRElt;
    setExtElt_t setExtElt = methods->characterSetElt;
    return defaultCommonVectorSubassign(storage, methods, s, nx, value,
                                        getRElt, setExtElt, warn);
}

static const SEXPREC*
defaultLogicalVectorSubassign(SEXP storage, storageMethods_t* methods,
                              const SEXPREC* s, int nx, const SEXPREC* value,
                              int* warn)
{
    getRElt_t getRElt = methods->logicalGetRElt;
    setExtElt_t setExtElt = methods->logicalSetElt;
    return defaultCommonVectorSubassign(storage, methods, s, nx, value,
                                        getRElt, setExtElt, warn);
}

static const SEXPREC*
defaultAnyVectorSubassign(SEXP storage, storageMethods_t* methods,
                          const SEXPREC* s, int nx, const SEXPREC* value,
                          int* warn)
{
    getRElt_t getRElt = methods->listGetRElt;
    setExtElt_t setExtElt = methods->anySetElt;
    return defaultCommonVectorSubassign(storage, methods, s, nx, value,
                                        getRElt, setExtElt, warn);
}

static SEXP
defaultCommonMatrixSubset(const SEXPREC* storage, storageMethods_t* methods,
                          const SEXPREC* s1,
                          const SEXPREC* s2, int m, int n,
                          SEXP value, SEXPTYPE mode, setRElt_t setRElt,
                          getExtMatrixElt_t getExtElt,
                          primitiveElt_t* NA_TYPE, int* warn)
{
    primitiveElt_t elt;
    const SEXPREC* s1type = getIndexType(s1);
    const SEXPREC* s2type = getIndexType(s2);
    if (s1type == noneIndex || s2type == noneIndex) {
        if (value == R_NilValue) {
            int ns1, ns2;
            if (s1type == noneIndex) {
                ns1 = 0;
                if (s2type == noneIndex)
                    ns2 = 0;
                else if (s2type == allIndex)
                    ns2 = n;
                else if (s2type == NAIndex)
                    ns2 = asInteger(GET_SLOT((SEXP)s2, install("length")));
                else if (s2type == sliceIndex)
                    ns2 = (int) REAL(GET_SLOT((SEXP)s2,
                                              install("content")))[1];
                else ns2 = LENGTH((SEXP) s2);
            } else {
                ns2 = 0;
                if (s1type == allIndex)
                    ns1 = m;
                else if (s1type == NAIndex)
                    ns1 = asInteger(GET_SLOT((SEXP)s1, install("length")));
                else if (s1type == sliceIndex)
                    ns1 = (int) REAL(GET_SLOT((SEXP)s1, install("content")))[1];
                else ns1 = LENGTH((SEXP) s1);
            }
            value = allocMatrix(mode, ns1, ns2);
        }
        return value;
    } else if (s1type == NAIndex || s2type == NAIndex) {
        int ns1, ns2, i;
        if (value == R_NilValue) {
            if (s1type == NAIndex) {
                ns1 = asInteger(GET_SLOT((SEXP)s1, install("length")));
                if (s2type == noneIndex)
                    ns2 = 0;
                else if (s2type == allIndex)
                    ns2 = n;
                else if (s2type == NAIndex)
                    ns2 = asInteger(GET_SLOT((SEXP)s2, install("length")));
                else if (s2type == sliceIndex)
                    ns2 = (int) REAL(GET_SLOT((SEXP)s2, install("content")))[1];
                else ns2 = LENGTH((SEXP) s2);
            } else {
                ns2 = asInteger(GET_SLOT((SEXP)s2, install("length")));
                if (s1type == allIndex)
                    ns1 = m;
                else if (s1type == noneIndex)
                    ns1 = 0;
                else if (s1type == sliceIndex)
                    ns1 = (int) REAL(GET_SLOT((SEXP)s1, install("content")))[1];
                else ns1 = LENGTH((SEXP) s1);
            }
            value = allocMatrix(mode, ns1, ns2);
        } else {
            int* dim = INTEGER(GET_DIM(value));
            ns1 = dim[0];
            ns2 = dim[1];
        }
        for (i = 0; i < ns1; i++) {
            int j;
            for (j = 0; j < ns2; j++)
                setRElt(value, i+j*ns1, NA_TYPE);
        }
        return value;
    } else if (s2type == allIndex) {
        int ns2 = n;
        int tmp2 = n;
        int i, j;
        /*
         * This assumes that ns2 has been set before calling the macro.
         * Also, the OUTER_LOOP_START macro must set jj and j2.
         */
#define INNER_SWITCH                                                    \
        do {                                                            \
            if (s1type == allIndex) {                                   \
                int ns1 = m;                                            \
                if (value == R_NilValue) {                              \
                    value = allocMatrix(mode, m, ns2);                  \
                }                                                       \
                PROTECT(value);                                         \
                OUTER_LOOP_START;                                       \
                if (j >= tmp2) {                                        \
                    for (i = 0; i < m; i++) {                           \
                        setRElt(value, i + jj, NA_TYPE);                \
                    }                                                   \
                } else {                                                \
                    for (i = 0; i < m; i++) {                           \
                        getExtElt(storage, methods, i, j2, m, n,        \
                                  &elt, warn);                          \
                        setRElt(value, i + jj, &elt);                   \
                    }                                                   \
                }                                                       \
                OUTER_LOOP_END;                                         \
            } else if (s1type == sliceIndex) {                          \
                double* content1 = REAL(GET_SLOT((SEXP)s1,              \
                                                 install("content")));  \
                int start1 = (int) content1[0];                         \
                int ns1 = (int) content1[1];                            \
                int stride1 = (int) content1[2];                        \
                int tmp1;                                               \
                int i1;                                                 \
                if (value == R_NilValue) {                              \
                    value = allocMatrix(mode, ns1, ns2);                \
                }                                                       \
                PROTECT(value);                                         \
                if (start1+(ns1-1)*stride1 >= m)                        \
                    tmp1 = 1+(m-start1)/stride1;                        \
                else tmp1 = ns1;                                        \
                OUTER_LOOP_START;                                       \
                if (j >= tmp2) {                                        \
                    for (i = 0; i < ns1; i++) {                         \
                        setRElt(value, i + jj, NA_TYPE);                \
                    }                                                   \
                } else {                                                \
                    for (i = 0, i1 = start1; i < tmp1;                  \
                         i++, i1 += stride1) {                          \
                        getExtElt(storage, methods, i1, j2,             \
                                  m, n, &elt, warn);                    \
                        setRElt(value, i+jj, &elt);                     \
                    }                                                   \
                    for (; i < ns1; i++, i1 += stride1) {               \
                        setRElt(value, i+jj, NA_TYPE);                  \
                    }                                                   \
                }                                                       \
                OUTER_LOOP_END;                                         \
            } else {                                                    \
                int* indx1 = INTEGER((SEXP) s1);                        \
                int ns1 = LENGTH((SEXP) s1);                            \
                if (value == R_NilValue) {                              \
                    value = allocMatrix(mode, ns1, ns2);                \
                }                                                       \
                PROTECT(value);                                         \
                OUTER_LOOP_START;                                       \
                if (j >= tmp2) {                                        \
                    for (i = 0; i < ns1; i++) {                         \
                        setRElt(value, i + jj, NA_TYPE);                \
                    }                                                   \
                } else {                                                \
                    for (i = 0; i < ns1; i++) {                         \
                        int i1 = indx1[i];                              \
                        if (i1 >= m || i1 == NA_INTEGER)                \
                            setRElt(value, i+jj, NA_TYPE);              \
                        else {                                          \
                            getExtElt(storage, methods, i1, j2,         \
                                      m, n, &elt, warn);                \
                            setRElt(value, i+jj, &elt);                 \
                        }                                               \
                    }                                                   \
                }                                                       \
                OUTER_LOOP_END;                                         \
            }                                                           \
        } while (0)

#define OUTER_LOOP_START                        \
        do {                                    \
            for (j = 0; j < n; j++) {           \
                int jj = j*ns1;                   \
                int j2 = j
#define OUTER_LOOP_END                          \
            }                                   \
        } while (0)

        INNER_SWITCH;
    } else if (s2type == sliceIndex) {
        double* content2 = REAL(GET_SLOT((SEXP)s2, install("content")));
        int start2 = (int) content2[0];
        int ns2 = (int) content2[1];
        int stride2 = (int) content2[2];
        int tmp2;
        int i, j;

        if (start2+(ns2-1)*stride2 >= n)
            tmp2 = 1+(n-start2)/stride2;
        else tmp2 = ns2;

#undef OUTER_LOOP_START
#undef OUTER_LOOP_END
#define OUTER_LOOP_START                               \
        do {                                           \
            int j2 = start2;                           \
            for (j = 0; j < ns2; j++, j2 += stride2) { \
                int jj = j*ns1
#define OUTER_LOOP_END                          \
            }                                   \
        } while (0)

        INNER_SWITCH;
    } else {
        int* indx2 = INTEGER((SEXP) s2);
        int ns2 = LENGTH((SEXP) s2);
        int tmp2;
        int i, j;
        if (ns2 > n)
            tmp2 = n;
        else tmp2 = ns2;

#undef OUTER_LOOP_START
#undef OUTER_LOOP_END
#define OUTER_LOOP_START                        \
        do {                                    \
            for (j = 0; j < ns2; j++) {         \
                int jj = j*ns1;                 \
                int j2 = indx2[j]
#define OUTER_LOOP_END                          \
            }                                   \
        } while (0)

        INNER_SWITCH;
    }

#undef INNER_SWITCH
#undef OUTER_LOOP_START
#undef OUTER_LOOP_END
    UNPROTECT(1);
    return value;
}

static SEXP
defaultMatrixSubset(const SEXPREC* storage, storageMethods_t* methods,
                    const SEXPREC* s1,
                    const SEXPREC* s2, int m, int n,
                    SEXP value, int* warn)
{
    setRElt_t setRElt = methods->setRElt;
    getExtMatrixElt_t getExtElt = methods->getMatrixElt;
    primitiveElt_t* NA_TYPE = &(methods->NA);

    return defaultCommonMatrixSubset(storage, methods, s1, s2,
                                     m, n, value, methods->mode, setRElt,
                                     getExtElt, NA_TYPE, warn);
}

static SEXP
defaultLogicalMatrixSubset(const SEXPREC* storage, storageMethods_t* methods,
                           const SEXPREC* s1,
                           const SEXPREC* s2, int m, int n,
                           SEXP value, int* warn)
{
    setRElt_t setRElt = methods->logicalSetRElt;
    getExtMatrixElt_t getExtElt = methods->logicalGetMatrixElt;
    primitiveElt_t NA;

    NA.logical = NA_LOGICAL;
    return defaultCommonMatrixSubset(storage, methods, s1, s2,
                                     m, n, value, LGLSXP, setRElt,
                                     getExtElt, &NA, warn);
}

static SEXP
defaultIntegerMatrixSubset(const SEXPREC* storage, storageMethods_t* methods,
                           const SEXPREC* s1,
                           const SEXPREC* s2, int m, int n,
                           SEXP value, int* warn)
{
    setRElt_t setRElt = methods->integerSetRElt;
    getExtMatrixElt_t getExtElt = methods->integerGetMatrixElt;
    primitiveElt_t NA;

    NA.integer = NA_INTEGER;
    return defaultCommonMatrixSubset(storage, methods, s1, s2,
                                     m, n, value, INTSXP, setRElt,
                                     getExtElt, &NA, warn);
}

static SEXP
defaultNumericMatrixSubset(const SEXPREC* storage, storageMethods_t* methods,
                           const SEXPREC* s1,
                           const SEXPREC* s2, int m, int n,
                           SEXP value, int* warn)
{
    setRElt_t setRElt = methods->numericSetRElt;
    getExtMatrixElt_t getExtElt = methods->numericGetMatrixElt;
    primitiveElt_t NA;

    NA.numeric = NA_REAL;
    return defaultCommonMatrixSubset(storage, methods, s1, s2,
                                     m, n, value, REALSXP, setRElt,
                                     getExtElt, &NA, warn);
}

static SEXP
defaultComplexMatrixSubset(const SEXPREC* storage, storageMethods_t* methods,
                           const SEXPREC* s1,
                           const SEXPREC* s2, int m, int n,
                           SEXP value, int* warn)
{
    setRElt_t setRElt = methods->complexSetRElt;
    getExtMatrixElt_t getExtElt = methods->complexGetMatrixElt;
    primitiveElt_t NA;

    NA.cmplx.r = NA.cmplx.i = NA_REAL;
    return defaultCommonMatrixSubset(storage, methods, s1, s2,
                                     m, n, value, CPLXSXP, setRElt,
                                     getExtElt, &NA, warn);
}

static SEXP
defaultCharacterMatrixSubset(const SEXPREC* storage, storageMethods_t* methods,
                             const SEXPREC* s1,
                             const SEXPREC* s2, int m, int n,
                             SEXP value, int* warn)
{
    setRElt_t setRElt = methods->characterSetRElt;
    getExtMatrixElt_t getExtElt = methods->characterGetMatrixElt;
    primitiveElt_t NA;

    NA.character = NA_STRING;
    return defaultCommonMatrixSubset(storage, methods, s1, s2,
                                     m, n, value, STRSXP, setRElt,
                                     getExtElt, &NA, warn);
}

static SEXP
defaultAnyMatrixSubset(const SEXPREC* storage, storageMethods_t* methods,
                             const SEXPREC* s1,
                             const SEXPREC* s2, int m, int n,
                             SEXP value, int* warn)
{
    setRElt_t setRElt = methods->listSetRElt;
    getExtMatrixElt_t getExtElt = methods->anyGetMatrixElt;
    primitiveElt_t NA;

    NA.any = R_NilValue;
    return defaultCommonMatrixSubset(storage, methods, s1, s2,
                                     m, n, value, VECSXP, setRElt,
                                     getExtElt, &NA, warn);
}

static const SEXPREC*
defaultCommonMatrixSubassign(SEXP storage, storageMethods_t* methods,
                             const SEXPREC* s1,
                             const SEXPREC* s2, int m, int n,
                             const SEXPREC* value, getRElt_t getRElt,
                             setExtMatrixElt_t setExtElt, int* warn)
{
    primitiveElt_t elt;
    const SEXPREC* s1type = getIndexType(s1);
    const SEXPREC* s2type = getIndexType(s2);
    int nvalue = LENGTH((SEXP)value);
    int k = 0;
    if (nvalue == 0 || s1type == noneIndex || s2type == noneIndex ||
        s1type == NAIndex || s2type == NAIndex) {
        return value;
    } else if (s2type == allIndex) {
        int i, j;
        /*
         * This assumes that ns2 has been set before calling the macro.
         * Also, the OUTER_LOOP_START macro must set jj and j2.
         */
#define INNER_SWITCH                                                    \
        do {                                                            \
            if (s1type == allIndex) {                                   \
                OUTER_LOOP_START;                                       \
                for (i = 0; i < m; i++) {                               \
                    getRElt(value, k++, &elt);                          \
                    setExtElt(storage, methods, i, j2, m, n,            \
                              &elt, warn);                              \
                    if (k == nvalue)                                    \
                        k = 0;                                          \
                }                                                       \
                OUTER_LOOP_END;                                         \
            } else if (s1type == sliceIndex) {                          \
                double* content1 = REAL(GET_SLOT((SEXP)s1,              \
                                                 install("content")));  \
                int start1 = (int) content1[0];                         \
                int ns1 = (int) content1[1];                            \
                int stride1 = (int) content1[2];                        \
                int i1;                                                 \
                OUTER_LOOP_START;                                       \
                for (i = 0, i1 = start1; i < ns1;                       \
                     i++, i1 += stride1) {                              \
                    getRElt(value, k++, &elt);                          \
                    setExtElt(storage, methods, i1, j2,                 \
                              m, n, &elt, warn);                        \
                    if (k == nvalue)                                    \
                        k = 0;                                          \
                }                                                       \
                OUTER_LOOP_END;                                         \
            } else {                                                    \
                int* indx1 = INTEGER((SEXP) s1);                        \
                int ns1 = LENGTH((SEXP) s1);                            \
                OUTER_LOOP_START;                                       \
                for (i = 0; i < ns1; i++) {                             \
                    int i1 = indx1[i];                                  \
                    getRElt(value, k++, &elt);                          \
                    setExtElt(storage, methods, i1, j2,                 \
                              m, n, &elt, warn);                        \
                    if (k == nvalue)                                    \
                        k = 0;                                          \
                }                                                       \
                OUTER_LOOP_END;                                         \
            }                                                           \
        } while (0)

#define OUTER_LOOP_START                        \
        do {                                    \
            for (j = 0; j < n; j++) {           \
                int j2 = j
#define OUTER_LOOP_END                          \
            }                                   \
        } while (0)

        INNER_SWITCH;
    } else if (s2type == sliceIndex) {
        double* content2 = REAL(GET_SLOT((SEXP)s2, install("content")));
        int start2 = (int) content2[0];
        int ns2 = (int) content2[1];
        int stride2 = (int) content2[2];
        int i, j;

#undef OUTER_LOOP_START
#undef OUTER_LOOP_END
#define OUTER_LOOP_START                               \
        do {                                           \
            int j2 = start2;                           \
            for (j = 0; j < ns2; j++, j2 += stride2) { \
                
#define OUTER_LOOP_END                          \
            }                                   \
        } while (0)

        INNER_SWITCH;
    } else {
        int* indx2 = INTEGER((SEXP) s2);
        int ns2 = LENGTH((SEXP) s2);
        int i, j;

#undef OUTER_LOOP_START
#undef OUTER_LOOP_END
#define OUTER_LOOP_START                        \
        do {                                    \
            for (j = 0; j < ns2; j++) {         \
                int j2 = indx2[j]
#define OUTER_LOOP_END                          \
            }                                   \
        } while (0)

        INNER_SWITCH;
    }

#undef INNER_IF
#undef OUTER_LOOP_START
#undef OUTER_LOOP_END

    return value;
}

static const SEXPREC*
defaultMatrixSubassign(SEXP storage, storageMethods_t* methods,
                       const SEXPREC* s1,
                       const SEXPREC* s2, int m, int n,
                       const SEXPREC* value,
                       int* warn)
{
    getRElt_t getRElt = methods->getRElt;
    setExtMatrixElt_t setExtElt = methods->setMatrixElt;

    return defaultCommonMatrixSubassign(storage, methods, s1,
                                        s2, m, n, value, getRElt,
                                        setExtElt, warn);
}

static const SEXPREC*
defaultLogicalMatrixSubassign(SEXP storage, storageMethods_t* methods,
                              const SEXPREC* s1,
                              const SEXPREC* s2, int m, int n,
                              const SEXPREC* value,
                              int* warn)
{
    getRElt_t getRElt = methods->logicalGetRElt;
    setExtMatrixElt_t setExtElt = methods->logicalSetMatrixElt;

    return defaultCommonMatrixSubassign(storage, methods, s1,
                                        s2, m, n, value, getRElt,
                                        setExtElt, warn);
}

static const SEXPREC*
defaultIntegerMatrixSubassign(SEXP storage, storageMethods_t* methods,
                              const SEXPREC* s1,
                              const SEXPREC* s2, int m, int n,
                              const SEXPREC* value,
                              int* warn)
{
    getRElt_t getRElt = methods->integerGetRElt;
    setExtMatrixElt_t setExtElt = methods->integerSetMatrixElt;

    return defaultCommonMatrixSubassign(storage, methods, s1,
                                        s2, m, n, value, getRElt,
                                        setExtElt, warn);
}

static const SEXPREC*
defaultNumericMatrixSubassign(SEXP storage, storageMethods_t* methods,
                              const SEXPREC* s1,
                              const SEXPREC* s2, int m, int n,
                              const SEXPREC* value,
                              int* warn)
{
    getRElt_t getRElt = methods->numericGetRElt;
    setExtMatrixElt_t setExtElt = methods->numericSetMatrixElt;

    return defaultCommonMatrixSubassign(storage, methods, s1,
                                        s2, m, n, value, getRElt,
                                        setExtElt, warn);
}

static const SEXPREC*
defaultComplexMatrixSubassign(SEXP storage, storageMethods_t* methods,
                              const SEXPREC* s1,
                              const SEXPREC* s2, int m, int n,
                              const SEXPREC* value,
                              int* warn)
{
    getRElt_t getRElt = methods->complexGetRElt;
    setExtMatrixElt_t setExtElt = methods->complexSetMatrixElt;

    return defaultCommonMatrixSubassign(storage, methods, s1,
                                        s2, m, n, value, getRElt,
                                        setExtElt, warn);
}

static const SEXPREC*
defaultCharacterMatrixSubassign(SEXP storage, storageMethods_t* methods,
                                const SEXPREC* s1,
                                const SEXPREC* s2, int m, int n,
                                const SEXPREC* value,
                                int* warn)
{
    getRElt_t getRElt = methods->characterGetRElt;
    setExtMatrixElt_t setExtElt = methods->characterSetMatrixElt;

    return defaultCommonMatrixSubassign(storage, methods, s1,
                                        s2, m, n, value, getRElt,
                                        setExtElt, warn);
}

static const SEXPREC*
defaultAnyMatrixSubassign(SEXP storage, storageMethods_t* methods,
                          const SEXPREC* s1,
                          const SEXPREC* s2, int m, int n,
                          const SEXPREC* value,
                          int* warn)
{
    getRElt_t getRElt = methods->listGetRElt;
    setExtMatrixElt_t setExtElt = methods->anySetMatrixElt;

    return defaultCommonMatrixSubassign(storage, methods, s1,
                                        s2, m, n, value, getRElt,
                                        setExtElt, warn);
}

static int
defaultSize(const SEXPREC* x)
{
    error("no size information");
    return 0; /* -Wall */
}

static SEXP
defaultResize(SEXP x, int n)
{
    error("can not resize this class of storage objects");
    return R_NilValue; /* -Wall */
}

static SEXP
storageMethodsStructSize()
{
    return ScalarInteger(sizeof(storageMethods_t));
}

static primitiveElt_t*
logicalGetRElt(const SEXPREC* x, int i,
               primitiveElt_t* val)
{
    val->logical = LOGICAL((SEXP) x)[i];
    return val;
}

static const primitiveElt_t*
logicalSetRElt(SEXP x, int i, const primitiveElt_t* val)
{
    LOGICAL(x)[i] = val->logical;
    return val;
}

static primitiveElt_t*
integerGetRElt(const SEXPREC* x, int i,
               primitiveElt_t* val)
{
    val->integer = INTEGER((SEXP) x)[i];
    return val;
}

static const primitiveElt_t*
integerSetRElt(SEXP x, int i, const primitiveElt_t* val)
{
    INTEGER(x)[i] = val->integer;
    return val;
}

static primitiveElt_t*
numericGetRElt(const SEXPREC* x, int i,
               primitiveElt_t* val)
{
    val->numeric = REAL((SEXP) x)[i];
    return val;
}

static const primitiveElt_t*
numericSetRElt(SEXP x, int i, const primitiveElt_t* val)
{
    REAL(x)[i] = val->numeric;
    return val;
}

static primitiveElt_t*
complexGetRElt(const SEXPREC* x, int i,
               primitiveElt_t* val)
{
    val->cmplx.r = COMPLEX((SEXP) x)[i].r;
    val->cmplx.i = COMPLEX((SEXP) x)[i].i;
    return val;
}

static const primitiveElt_t*
complexSetRElt(SEXP x, int i, const primitiveElt_t* val)
{
    COMPLEX(x)[i].r = val->cmplx.r;
    COMPLEX(x)[i].i = val->cmplx.i;
    return val;
}

static primitiveElt_t*
characterGetRElt(const SEXPREC* x, int i,
               primitiveElt_t* val)
{
    val->character = STRING_ELT((SEXP) x, i);
    return val;
}

static const primitiveElt_t*
characterSetRElt(SEXP x, int i, const primitiveElt_t* val)
{
    SET_STRING_ELT(x, i, (SEXP) val->character);
    return val;
}

static primitiveElt_t*
listGetRElt(const SEXPREC* x, int i,
            primitiveElt_t* val)
{
    val->any = VECTOR_ELT((SEXP) x, i);
    return val;
}

static const primitiveElt_t*
listSetRElt(SEXP x, int i, const primitiveElt_t* val)
{
    SET_VECTOR_ELT(x, i, (SEXP) val->any);
    return val;
}

enum {GETELT_METHOD,
      SETELT_METHOD,
      VECSUBSET_METHOD,
      VECSUBASSIGN_METHOD,
      MATGETELT_METHOD,
      MATSETELT_METHOD,
      MATSUBSET_METHOD,
      MATSUBASSIGN_METHOD,
      SIZE_METHOD,
      RESIZE_METHOD};

static int storageMethodsNameToIndex(const char* nm)
{
    int n;

    if (strcmp(nm, "size") == 0)
        return SIZE_METHOD;
    if (strcmp(nm, "resize") == 0)
        return RESIZE_METHOD;
    n = strlen(nm);
    if (n > 7) {
        if (strncmp(nm, "imaginary", 9) == 0)
            nm += 9;
        else if (strncmp(nm, "character", 9) == 0)
            nm += 9;
        else if (strncmp(nm, "any", 3) == 0)
            nm += 3;
        else nm += 7;
        if (strcmp(nm, "MatrixSubset") == 0)
            return MATSUBSET_METHOD;
        if (strcmp(nm, "MatrixSubassign") == 0)
            return MATSUBASSIGN_METHOD;
        if (strcmp(nm, "GetElt") == 0)
            return GETELT_METHOD;
        if (strcmp(nm, "SetElt") == 0)
            return SETELT_METHOD;
        if (strcmp(nm, "Subset") == 0)
            return VECSUBSET_METHOD;
        if (strcmp(nm, "Subassign") == 0)
            return VECSUBASSIGN_METHOD;
        if (strcmp(nm, "GetMatrixElt") == 0)
            return MATGETELT_METHOD;
        if (strcmp(nm, "SetMatrixElt") == 0)
            return MATSETELT_METHOD;
    }
    return -1;
}

static int
storageMethodsNameToMode(const char* nm)
{
    char buf[10];
    int n = strlen(nm);
    if (n > 9)
        n = 9;
    memcpy(buf, nm, n);
    if (n == 9) {
        buf[9] = '\0';
        if (strcmp(buf, "character") == 0)
            return STRSXP;
        if (strcmp(buf, "imaginary") == 0)
            return -REALSXP;
    }
    if (n >= 7) {
        buf[7] = '\0';
        if (strcmp(buf, "numeric") == 0)
            return REALSXP;
        if (strcmp(buf, "integer") == 0)
            return INTSXP;
        if (strcmp(buf, "logical") == 0)
            return LGLSXP;
        if (strcmp(buf, "complex") == 0)
            return CPLXSXP;
    }
    if (n >= 3) {
        buf[3] = '\0';
        if (strcmp(buf, "any") == 0)
            return VECSXP;
    }
    return -1;
}

static SEXP defaultAlloc(const SEXPREC* x, SEXP type, int len)
{
    SEXP klass = GET_CLASS((SEXP)x);
    SEXP call = PROTECT(lang4(install("new"), klass,
                              type, ScalarInteger(len)));
    SEXP ans;

    SET_TAG(CDDR(call), install("type"));
    SET_TAG(CDR(CDDR(call)), install("length"));
    ans = eval(call, R_GlobalEnv);
    UNPROTECT(1);
    return ans;
}

static SEXP
initNativeStorageMethods(SEXP methods)
{
    storageMethods_t* allMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(methods,
                                            install("allMethods")),
                                   install("ptr")));
    storageMethods_t* numericMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(methods,
                                            install("numericMethods")),
                                   install("ptr")));
    storageMethods_t* integerMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(methods,
                                            install("integerMethods")),
                                   install("ptr")));
    storageMethods_t* logicalMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(methods,
                                            install("logicalMethods")),
                                   install("ptr")));
    storageMethods_t* complexMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(methods,
                                            install("complexMethods")),
                                   install("ptr")));
    storageMethods_t* characterMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(methods,
                                            install("characterMethods")),
                                   install("ptr")));
    storageMethods_t* listMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(methods,
                                            install("listMethods")),
                                   install("ptr")));

    allMethods->logicalGetRElt = logicalMethods->logicalGetRElt =
        integerMethods->logicalGetRElt = numericMethods->logicalGetRElt =
        complexMethods->logicalGetRElt = characterMethods->logicalGetRElt =
        listMethods->logicalGetRElt = logicalMethods->getRElt = &logicalGetRElt;
    allMethods->integerGetRElt = logicalMethods->integerGetRElt =
        integerMethods->integerGetRElt = numericMethods->integerGetRElt =
        complexMethods->integerGetRElt = characterMethods->integerGetRElt =
        listMethods->integerGetRElt = integerMethods->getRElt = &integerGetRElt;
    allMethods->numericGetRElt = logicalMethods->numericGetRElt =
        integerMethods->numericGetRElt = numericMethods->numericGetRElt =
        complexMethods->numericGetRElt = characterMethods->numericGetRElt =
        listMethods->numericGetRElt = numericMethods->getRElt = &numericGetRElt;
    allMethods->complexGetRElt = logicalMethods->complexGetRElt =
        integerMethods->complexGetRElt = numericMethods->complexGetRElt =
        complexMethods->complexGetRElt = characterMethods->complexGetRElt =
        listMethods->complexGetRElt = complexMethods->getRElt = &complexGetRElt;
    allMethods->characterGetRElt = logicalMethods->characterGetRElt =
        integerMethods->characterGetRElt = numericMethods->characterGetRElt =
        complexMethods->characterGetRElt = characterMethods->characterGetRElt =
        listMethods->characterGetRElt = characterMethods->getRElt =
        &characterGetRElt;
    allMethods->listGetRElt = logicalMethods->listGetRElt =
        integerMethods->listGetRElt = numericMethods->listGetRElt =
        complexMethods->listGetRElt = characterMethods->listGetRElt =
        listMethods->listGetRElt = listMethods->getRElt = &listGetRElt;

    allMethods->logicalSetRElt = logicalMethods->logicalSetRElt =
        integerMethods->logicalSetRElt = numericMethods->logicalSetRElt =
        complexMethods->logicalSetRElt = characterMethods->logicalSetRElt =
        listMethods->logicalSetRElt = logicalMethods->setRElt = &logicalSetRElt;
    allMethods->integerSetRElt = logicalMethods->integerSetRElt =
        integerMethods->integerSetRElt = numericMethods->integerSetRElt =
        complexMethods->integerSetRElt = characterMethods->integerSetRElt =
        listMethods->integerSetRElt = integerMethods->setRElt = &integerSetRElt;
    allMethods->numericSetRElt = logicalMethods->numericSetRElt =
        integerMethods->numericSetRElt = numericMethods->numericSetRElt =
        complexMethods->numericSetRElt = characterMethods->numericSetRElt =
        listMethods->numericSetRElt = numericMethods->setRElt = &numericSetRElt;
    allMethods->complexSetRElt = logicalMethods->complexSetRElt =
        integerMethods->complexSetRElt = numericMethods->complexSetRElt =
        complexMethods->complexSetRElt = characterMethods->complexSetRElt =
        listMethods->complexSetRElt = complexMethods->setRElt = &complexSetRElt;
    allMethods->characterSetRElt = logicalMethods->characterSetRElt =
        integerMethods->characterSetRElt = numericMethods->characterSetRElt =
        complexMethods->characterSetRElt = characterMethods->characterSetRElt =
        listMethods->characterSetRElt = characterMethods->setRElt =
        &characterSetRElt;
    allMethods->listSetRElt = logicalMethods->listSetRElt =
        integerMethods->listSetRElt = numericMethods->listSetRElt =
        complexMethods->listSetRElt = characterMethods->listSetRElt =
        listMethods->listSetRElt = listMethods->setRElt = &listSetRElt;

    allMethods->numericGetElt = allMethods->integerGetElt =
        allMethods->logicalGetElt = allMethods->complexGetElt =
        allMethods->characterGetElt = allMethods->anyGetElt =
        allMethods->getExtElt = &defaultGetExtElt;
    numericMethods->getExtElt = integerMethods->getExtElt =
        logicalMethods->getExtElt = complexMethods->getExtElt =
        characterMethods->getExtElt = listMethods->getExtElt =
        &defaultGetExtElt;

    allMethods->imaginaryGetElt = complexMethods->imaginaryGetElt =
        &defaultGetImaginaryElt;
    logicalMethods->imaginaryGetElt =
        integerMethods->imaginaryGetElt = numericMethods->imaginaryGetElt =
        characterMethods->imaginaryGetElt = listMethods->imaginaryGetElt =
        &defaultGetImaginaryFromAnyElt;

    logicalMethods->integerGetElt = &defaultGetExtElt;
    logicalMethods->numericGetElt = &defaultGetNumericFromLogicalElt;
    logicalMethods->complexGetElt = &defaultGetComplexFromLogicalElt;
    logicalMethods->characterGetElt = &defaultGetCharacterFromLogicalElt;
    logicalMethods->anyGetElt = &defaultGetAnyFromLogicalElt;

    integerMethods->logicalGetElt = &defaultGetLogicalFromIntegerElt;
    integerMethods->numericGetElt = &defaultGetNumericFromIntegerElt;
    integerMethods->complexGetElt = &defaultGetComplexFromIntegerElt;
    integerMethods->characterGetElt = &defaultGetCharacterFromIntegerElt;
    integerMethods->anyGetElt = &defaultGetAnyFromIntegerElt;

    numericMethods->logicalGetElt = &defaultGetLogicalFromNumericElt;
    numericMethods->integerGetElt = &defaultGetIntegerFromNumericElt;
    numericMethods->complexGetElt = &defaultGetComplexFromNumericElt;
    numericMethods->characterGetElt = &defaultGetCharacterFromNumericElt;
    numericMethods->anyGetElt = &defaultGetAnyFromNumericElt;

    complexMethods->logicalGetElt = &defaultGetLogicalFromComplexElt;
    complexMethods->integerGetElt = &defaultGetIntegerFromComplexElt;
    complexMethods->numericGetElt = &defaultGetNumericFromComplexElt;
    complexMethods->characterGetElt = &defaultGetCharacterFromComplexElt;
    complexMethods->anyGetElt = &defaultGetAnyFromComplexElt;

    characterMethods->logicalGetElt = &defaultGetLogicalFromCharacterElt;
    characterMethods->integerGetElt = &defaultGetIntegerFromCharacterElt;
    characterMethods->numericGetElt = &defaultGetNumericFromCharacterElt;
    characterMethods->complexGetElt = &defaultGetComplexFromCharacterElt;
    characterMethods->anyGetElt = &defaultGetAnyFromCharacterElt;

    listMethods->logicalGetElt = &defaultGetLogicalFromAnyElt;
    listMethods->integerGetElt = &defaultGetIntegerFromAnyElt;
    listMethods->numericGetElt = &defaultGetNumericFromAnyElt;
    listMethods->complexGetElt = &defaultGetComplexFromAnyElt;
    listMethods->characterGetElt = &defaultGetCharacterFromAnyElt;

    allMethods->numericSetElt = allMethods->integerSetElt =
        allMethods->logicalSetElt = allMethods->complexSetElt =
        allMethods->characterSetElt = allMethods->anySetElt =
        allMethods->setExtElt = &defaultSetExtElt;
    numericMethods->setExtElt = integerMethods->setExtElt =
        logicalMethods->setExtElt = complexMethods->setExtElt =
        characterMethods->setExtElt = listMethods->setExtElt =
        &defaultSetExtElt;

    allMethods->imaginarySetElt = complexMethods->imaginarySetElt =
        &defaultSetImaginaryElt;
    logicalMethods->imaginarySetElt =
        integerMethods->imaginarySetElt = numericMethods->imaginarySetElt =
        characterMethods->imaginarySetElt = listMethods->imaginarySetElt =
        &defaultSetAnyToImaginaryElt;

    logicalMethods->integerSetElt = &defaultSetIntegerToLogicalElt;
    logicalMethods->numericSetElt = &defaultSetNumericToLogicalElt;
    logicalMethods->complexSetElt = &defaultSetComplexToLogicalElt;
    logicalMethods->characterSetElt = &defaultSetCharacterToLogicalElt;
    logicalMethods->anySetElt = &defaultSetAnyToLogicalElt;

    integerMethods->logicalSetElt = &defaultSetExtElt;
    integerMethods->numericSetElt = &defaultSetNumericToIntegerElt;
    integerMethods->complexSetElt = &defaultSetComplexToIntegerElt;
    integerMethods->characterSetElt = &defaultSetCharacterToIntegerElt;
    integerMethods->anySetElt = &defaultSetAnyToIntegerElt;
    
    numericMethods->logicalSetElt = &defaultSetLogicalToNumericElt;
    numericMethods->integerSetElt = &defaultSetIntegerToNumericElt;
    numericMethods->complexSetElt = &defaultSetComplexToNumericElt;
    numericMethods->characterSetElt = &defaultSetCharacterToNumericElt;
    numericMethods->anySetElt = &defaultSetAnyToNumericElt;

    complexMethods->logicalSetElt = &defaultSetLogicalToComplexElt;
    complexMethods->integerSetElt = &defaultSetIntegerToComplexElt;
    complexMethods->numericSetElt = &defaultSetNumericToComplexElt;
    complexMethods->characterSetElt = &defaultSetCharacterToComplexElt;
    complexMethods->anySetElt = &defaultSetAnyToComplexElt;

    characterMethods->logicalSetElt = &defaultSetLogicalToCharacterElt;
    characterMethods->integerSetElt = &defaultSetIntegerToCharacterElt;
    characterMethods->numericSetElt = &defaultSetNumericToCharacterElt;
    characterMethods->complexSetElt = &defaultSetComplexToCharacterElt;
    characterMethods->anySetElt = &defaultSetAnyToCharacterElt;

    listMethods->logicalSetElt = &defaultSetLogicalToAnyElt;
    listMethods->integerSetElt = &defaultSetIntegerToAnyElt;
    listMethods->numericSetElt = &defaultSetNumericToAnyElt;
    listMethods->complexSetElt = &defaultSetComplexToAnyElt;
    listMethods->characterSetElt = &defaultSetCharacterToAnyElt;

    allMethods->numericGetMatrixElt = allMethods->integerGetMatrixElt =
        allMethods->logicalGetMatrixElt = allMethods->complexGetMatrixElt =
        allMethods->characterGetMatrixElt = allMethods->anyGetMatrixElt =
        allMethods->getMatrixElt = &defaultGetExtMatrixElt;
    numericMethods->getMatrixElt = integerMethods->getMatrixElt =
        logicalMethods->getMatrixElt = complexMethods->getMatrixElt =
        characterMethods->getMatrixElt = listMethods->getMatrixElt =
        &defaultGetExtMatrixElt;

    allMethods->logicalGetMatrixElt =
        complexMethods->logicalGetMatrixElt =
        logicalMethods->logicalGetMatrixElt =
        integerMethods->logicalGetMatrixElt =
        numericMethods->logicalGetMatrixElt =
        characterMethods->logicalGetMatrixElt =
        listMethods->logicalGetMatrixElt =
        &defaultLogicalGetMatrixElt;

    allMethods->integerGetMatrixElt =
        complexMethods->integerGetMatrixElt =
        logicalMethods->integerGetMatrixElt =
        integerMethods->integerGetMatrixElt =
        numericMethods->integerGetMatrixElt =
        characterMethods->integerGetMatrixElt =
        listMethods->integerGetMatrixElt =
        &defaultIntegerGetMatrixElt;

    allMethods->numericGetMatrixElt =
        complexMethods->numericGetMatrixElt =
        logicalMethods->numericGetMatrixElt =
        integerMethods->numericGetMatrixElt =
        numericMethods->numericGetMatrixElt =
        characterMethods->numericGetMatrixElt =
        listMethods->numericGetMatrixElt =
        &defaultNumericGetMatrixElt;

    allMethods->complexGetMatrixElt =
        complexMethods->complexGetMatrixElt =
        logicalMethods->complexGetMatrixElt =
        integerMethods->complexGetMatrixElt =
        numericMethods->complexGetMatrixElt =
        characterMethods->complexGetMatrixElt =
        listMethods->complexGetMatrixElt =
        &defaultComplexGetMatrixElt;

    allMethods->characterGetMatrixElt =
        complexMethods->characterGetMatrixElt =
        logicalMethods->characterGetMatrixElt =
        integerMethods->characterGetMatrixElt =
        numericMethods->characterGetMatrixElt =
        characterMethods->characterGetMatrixElt =
        listMethods->characterGetMatrixElt =
        &defaultCharacterGetMatrixElt;

    allMethods->anyGetMatrixElt =
        complexMethods->anyGetMatrixElt =
        logicalMethods->anyGetMatrixElt =
        integerMethods->anyGetMatrixElt =
        numericMethods->anyGetMatrixElt =
        characterMethods->anyGetMatrixElt =
        listMethods->anyGetMatrixElt =
        &defaultAnyGetMatrixElt;

    allMethods->imaginaryGetMatrixElt =
        complexMethods->imaginaryGetMatrixElt =
        logicalMethods->imaginaryGetMatrixElt =
        integerMethods->imaginaryGetMatrixElt =
        numericMethods->imaginaryGetMatrixElt =
        characterMethods->imaginaryGetMatrixElt =
        listMethods->imaginaryGetMatrixElt =
        &defaultImaginaryGetMatrixElt;

    allMethods->numericSetMatrixElt = allMethods->integerSetMatrixElt =
        allMethods->logicalSetMatrixElt = allMethods->complexSetMatrixElt =
        allMethods->characterSetMatrixElt = allMethods->anySetMatrixElt =
        allMethods->setMatrixElt = &defaultSetExtMatrixElt;
    numericMethods->setMatrixElt = integerMethods->setMatrixElt =
        logicalMethods->setMatrixElt = complexMethods->setMatrixElt =
        characterMethods->setMatrixElt = listMethods->setMatrixElt =
        &defaultSetExtMatrixElt;

    allMethods->logicalSetMatrixElt =
        complexMethods->logicalSetMatrixElt =
        logicalMethods->logicalSetMatrixElt =
        integerMethods->logicalSetMatrixElt =
        numericMethods->logicalSetMatrixElt =
        characterMethods->logicalSetMatrixElt =
        listMethods->logicalSetMatrixElt =
        &defaultLogicalSetMatrixElt;

    allMethods->integerSetMatrixElt =
        complexMethods->integerSetMatrixElt =
        logicalMethods->integerSetMatrixElt =
        integerMethods->integerSetMatrixElt =
        numericMethods->integerSetMatrixElt =
        characterMethods->integerSetMatrixElt =
        listMethods->integerSetMatrixElt =
        &defaultIntegerSetMatrixElt;

    allMethods->numericSetMatrixElt =
        complexMethods->numericSetMatrixElt =
        logicalMethods->numericSetMatrixElt =
        integerMethods->numericSetMatrixElt =
        numericMethods->numericSetMatrixElt =
        characterMethods->numericSetMatrixElt =
        listMethods->numericSetMatrixElt =
        &defaultNumericSetMatrixElt;

    allMethods->complexSetMatrixElt =
        complexMethods->complexSetMatrixElt =
        logicalMethods->complexSetMatrixElt =
        integerMethods->complexSetMatrixElt =
        numericMethods->complexSetMatrixElt =
        characterMethods->complexSetMatrixElt =
        listMethods->complexSetMatrixElt =
        &defaultComplexSetMatrixElt;

    allMethods->characterSetMatrixElt =
        complexMethods->characterSetMatrixElt =
        logicalMethods->characterSetMatrixElt =
        integerMethods->characterSetMatrixElt =
        numericMethods->characterSetMatrixElt =
        characterMethods->characterSetMatrixElt =
        listMethods->characterSetMatrixElt =
        &defaultCharacterSetMatrixElt;

    allMethods->anySetMatrixElt =
        complexMethods->anySetMatrixElt =
        logicalMethods->anySetMatrixElt =
        integerMethods->anySetMatrixElt =
        numericMethods->anySetMatrixElt =
        characterMethods->anySetMatrixElt =
        listMethods->anySetMatrixElt =
        &defaultAnySetMatrixElt;

    allMethods->imaginarySetMatrixElt =
        complexMethods->imaginarySetMatrixElt =
        logicalMethods->imaginarySetMatrixElt =
        integerMethods->imaginarySetMatrixElt =
        numericMethods->imaginarySetMatrixElt =
        characterMethods->imaginarySetMatrixElt =
        listMethods->imaginarySetMatrixElt =
        &defaultImaginarySetMatrixElt;

    allMethods->numericSubset = allMethods->integerSubset =
        allMethods->logicalSubset = allMethods->complexSubset =
        allMethods->characterSubset = allMethods->anySubset =
        allMethods->vectorSubset = &defaultVectorSubset;
    numericMethods->vectorSubset = integerMethods->vectorSubset =
        logicalMethods->vectorSubset = complexMethods->vectorSubset =
        characterMethods->vectorSubset = listMethods->vectorSubset =
        &defaultVectorSubset;
    numericMethods->numericSubset = integerMethods->integerSubset =
        logicalMethods->logicalSubset = complexMethods->complexSubset =
        characterMethods->characterSubset = listMethods->anySubset =
        &defaultVectorSubset;
    numericMethods->logicalSubset = integerMethods->logicalSubset =
        characterMethods->logicalSubset = complexMethods->logicalSubset =
        listMethods->logicalSubset = &defaultLogicalVectorSubset;
    numericMethods->integerSubset = characterMethods->integerSubset =
        logicalMethods->integerSubset = complexMethods->integerSubset =
        listMethods->integerSubset = &defaultIntegerVectorSubset;
    characterMethods->numericSubset = integerMethods->numericSubset =
        logicalMethods->numericSubset = complexMethods->numericSubset =
        listMethods->numericSubset = &defaultNumericVectorSubset;
    numericMethods->complexSubset = integerMethods->complexSubset =
        logicalMethods->complexSubset = characterMethods->complexSubset =
        listMethods->complexSubset = &defaultComplexVectorSubset;
    numericMethods->characterSubset = integerMethods->characterSubset =
        logicalMethods->characterSubset = complexMethods->characterSubset =
        listMethods->characterSubset = &defaultCharacterVectorSubset;
    numericMethods->anySubset = integerMethods->anySubset =
        logicalMethods->anySubset = complexMethods->anySubset =
        characterMethods->anySubset = &defaultAnyVectorSubset;

    allMethods->numericSubassign = allMethods->integerSubassign =
        allMethods->logicalSubassign = allMethods->complexSubassign =
        allMethods->characterSubassign = allMethods->anySubassign =
        &defaultVectorSubassign;
    numericMethods->numericSubassign = integerMethods->integerSubassign =
        logicalMethods->logicalSubassign = complexMethods->complexSubassign =
        characterMethods->characterSubassign = listMethods->anySubassign =
        &defaultVectorSubassign;
    numericMethods->logicalSubassign = integerMethods->logicalSubassign =
        characterMethods->logicalSubassign = complexMethods->logicalSubassign =
        listMethods->logicalSubassign = &defaultLogicalVectorSubassign;
    numericMethods->integerSubassign = characterMethods->integerSubassign =
        logicalMethods->integerSubassign = complexMethods->integerSubassign =
        listMethods->integerSubassign = &defaultIntegerVectorSubassign;
    characterMethods->numericSubassign = integerMethods->numericSubassign =
        logicalMethods->numericSubassign = complexMethods->numericSubassign =
        listMethods->numericSubassign = &defaultNumericVectorSubassign;
    numericMethods->complexSubassign = integerMethods->complexSubassign =
        logicalMethods->complexSubassign = characterMethods->complexSubassign =
        listMethods->complexSubassign = &defaultComplexVectorSubassign;
    numericMethods->characterSubassign = integerMethods->characterSubassign =
        logicalMethods->characterSubassign =
        complexMethods->characterSubassign =
        listMethods->characterSubassign =
        &defaultCharacterVectorSubassign;
    numericMethods->anySubassign = integerMethods->anySubassign =
        logicalMethods->anySubassign = characterMethods->anySubassign =
        complexMethods->anySubassign = &defaultAnyVectorSubassign;

    allMethods->numericMatrixSubset = allMethods->integerMatrixSubset =
        allMethods->logicalMatrixSubset = allMethods->complexMatrixSubset =
        allMethods->characterMatrixSubset = allMethods->anyMatrixSubset =
        allMethods->matrixSubset = &defaultMatrixSubset;
    numericMethods->matrixSubset = integerMethods->matrixSubset =
        logicalMethods->matrixSubset = complexMethods->matrixSubset =
        characterMethods->matrixSubset = listMethods->matrixSubset =
        &defaultMatrixSubset;
    numericMethods->logicalMatrixSubset = integerMethods->logicalMatrixSubset =
        characterMethods->logicalMatrixSubset =
        complexMethods->logicalMatrixSubset =
        listMethods->logicalMatrixSubset =
        &defaultLogicalMatrixSubset;
    numericMethods->integerMatrixSubset =
        characterMethods->integerMatrixSubset =
        logicalMethods->integerMatrixSubset =
        complexMethods->integerMatrixSubset =
        listMethods->integerMatrixSubset =
        &defaultIntegerMatrixSubset;
    characterMethods->numericMatrixSubset =
        integerMethods->numericMatrixSubset =
        logicalMethods->numericMatrixSubset =
        complexMethods->numericMatrixSubset =
        listMethods->numericMatrixSubset =
        &defaultNumericMatrixSubset;
    numericMethods->complexMatrixSubset =
        integerMethods->complexMatrixSubset =
        logicalMethods->complexMatrixSubset =
        characterMethods->complexMatrixSubset =
        listMethods->complexMatrixSubset =
        &defaultComplexMatrixSubset;
    numericMethods->characterMatrixSubset =
        integerMethods->characterMatrixSubset =
        logicalMethods->characterMatrixSubset =
        complexMethods->characterMatrixSubset =
        listMethods->characterMatrixSubset =
        &defaultCharacterMatrixSubset;
    numericMethods->anyMatrixSubset =
        integerMethods->anyMatrixSubset =
        logicalMethods->anyMatrixSubset =
        complexMethods->anyMatrixSubset =
        characterMethods->anyMatrixSubset =
        &defaultAnyMatrixSubset;

    allMethods->numericMatrixSubassign = allMethods->integerMatrixSubassign =
        allMethods->logicalMatrixSubassign =
        allMethods->complexMatrixSubassign =
        allMethods->characterMatrixSubassign =
        allMethods->anyMatrixSubassign =
        &defaultMatrixSubassign;
    numericMethods->numericMatrixSubassign =
        integerMethods->integerMatrixSubassign =
        logicalMethods->logicalMatrixSubassign =
        complexMethods->complexMatrixSubassign =
        characterMethods->characterMatrixSubassign =
        listMethods->anyMatrixSubassign =
        &defaultMatrixSubassign;
    numericMethods->logicalMatrixSubassign =
        integerMethods->logicalMatrixSubassign =
        characterMethods->logicalMatrixSubassign =
        complexMethods->logicalMatrixSubassign =
        listMethods->logicalMatrixSubassign =
        &defaultLogicalMatrixSubassign;
    numericMethods->integerMatrixSubassign =
        characterMethods->integerMatrixSubassign =
        logicalMethods->integerMatrixSubassign =
        complexMethods->integerMatrixSubassign =
        listMethods->integerMatrixSubassign =
        &defaultIntegerMatrixSubassign;
    characterMethods->numericMatrixSubassign =
        integerMethods->numericMatrixSubassign =
        logicalMethods->numericMatrixSubassign =
        complexMethods->numericMatrixSubassign =
        listMethods->numericMatrixSubassign =
        &defaultNumericMatrixSubassign;
    numericMethods->complexMatrixSubassign =
        integerMethods->complexMatrixSubassign =
        logicalMethods->complexMatrixSubassign =
        characterMethods->complexMatrixSubassign =
        listMethods->complexMatrixSubassign =
        &defaultComplexMatrixSubassign;
    numericMethods->characterMatrixSubassign =
        integerMethods->characterMatrixSubassign =
        logicalMethods->characterMatrixSubassign =
        complexMethods->characterMatrixSubassign =
        listMethods->characterMatrixSubassign =
        &defaultCharacterMatrixSubassign;
    numericMethods->anyMatrixSubassign =
        integerMethods->anyMatrixSubassign =
        logicalMethods->anyMatrixSubassign =
        complexMethods->anyMatrixSubassign =
        characterMethods->anyMatrixSubassign =
        &defaultAnyMatrixSubassign;

    numericMethods->alloc = integerMethods->alloc =
        logicalMethods->alloc = complexMethods->alloc =
        characterMethods->alloc = listMethods->alloc =
        allMethods->alloc = &defaultAlloc;

    numericMethods->size = integerMethods->size =
        logicalMethods->size = complexMethods->size =
        characterMethods->size = listMethods->size =
        allMethods->size = &defaultSize;

    numericMethods->resize = integerMethods->resize =
        logicalMethods->resize = complexMethods->resize =
        characterMethods->resize = listMethods->resize =
        allMethods->resize = &defaultResize;

    logicalMethods->NA.logical = NA_LOGICAL;
    logicalMethods->mode = LGLSXP;
    integerMethods->NA.integer = NA_INTEGER;
    integerMethods->mode = INTSXP;
    numericMethods->NA.numeric = NA_REAL;
    numericMethods->mode = REALSXP;
    complexMethods->NA.cmplx.r = NA_REAL;
    complexMethods->NA.cmplx.i = NA_REAL;
    complexMethods->mode = CPLXSXP;
    characterMethods->NA.character = NA_STRING;
    characterMethods->mode = STRSXP;
    listMethods->NA.character = R_NilValue;
    listMethods->mode = VECSXP;

    return methods;
}

static SEXP
createNativeStorageMethods(SEXP thisMethods,
                           SEXP thisMethodsList, SEXP supMethods)
{
    int i;
    int n = LENGTH(thisMethodsList);
    SEXP names = GET_NAMES(thisMethodsList);

    storageMethods_t* allMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(thisMethods,
                                            install("allMethods")),
                                   install("ptr")));
    storageMethods_t* numericMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(thisMethods,
                                            install("numericMethods")),
                                   install("ptr")));
    storageMethods_t* integerMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(thisMethods,
                                            install("integerMethods")),
                                   install("ptr")));
    storageMethods_t* logicalMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(thisMethods,
                                            install("logicalMethods")),
                                   install("ptr")));
    storageMethods_t* complexMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(thisMethods,
                                            install("complexMethods")),
                                   install("ptr")));
    storageMethods_t* characterMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(thisMethods,
                                            install("characterMethods")),
                                   install("ptr")));
    storageMethods_t* listMethods =
        R_ExternalPtrAddr(GET_SLOT(GET_SLOT(thisMethods,
                                            install("listMethods")),
                                   install("ptr")));
    memcpy(allMethods,
           R_ExternalPtrAddr(GET_SLOT(GET_SLOT(supMethods,
                                               install("allMethods")),
                                      install("ptr"))),
           sizeof(storageMethods_t));
    memcpy(numericMethods,
           R_ExternalPtrAddr(GET_SLOT(GET_SLOT(supMethods,
                                               install("numericMethods")),
                                      install("ptr"))),
           sizeof(storageMethods_t));
    memcpy(integerMethods,
           R_ExternalPtrAddr(GET_SLOT(GET_SLOT(supMethods,
                                               install("integerMethods")),
                                      install("ptr"))),
           sizeof(storageMethods_t));
    memcpy(logicalMethods,
           R_ExternalPtrAddr(GET_SLOT(GET_SLOT(supMethods,
                                               install("logicalMethods")),
                                      install("ptr"))),
           sizeof(storageMethods_t));
    memcpy(complexMethods,
           R_ExternalPtrAddr(GET_SLOT(GET_SLOT(supMethods,
                                               install("complexMethods")),
                                      install("ptr"))),
           sizeof(storageMethods_t));
    memcpy(characterMethods,
           R_ExternalPtrAddr(GET_SLOT(GET_SLOT(supMethods,
                                               install("characterMethods")),
                                      install("ptr"))),
           sizeof(storageMethods_t));
    memcpy(listMethods,
           R_ExternalPtrAddr(GET_SLOT(GET_SLOT(supMethods,
                                               install("listMethods")),
                                      install("ptr"))),
           sizeof(storageMethods_t));

    for (i = 0; i < n; i++) {
        void* addr = R_ExternalPtrAddr(VECTOR_ELT(thisMethodsList, i));
        const char* nm = CHAR(STRING_ELT(names, i));
        int indx = storageMethodsNameToIndex(nm);
        SEXPTYPE mode = storageMethodsNameToMode(nm);

        switch (indx) {
        case SIZE_METHOD:
            allMethods->size = numericMethods->size =
                integerMethods->size = logicalMethods->size =
                complexMethods->size = characterMethods->size =
                listMethods->size = (int (*)(const SEXPREC*)) addr;
            break;
        case RESIZE_METHOD:
            allMethods->resize = numericMethods->resize =
                integerMethods->resize = logicalMethods->resize =
                complexMethods->resize = characterMethods->resize =
                listMethods->resize = (SEXP (*)(SEXP, int)) addr;
            break;
        case GETELT_METHOD:
            switch (mode) {
            case REALSXP:
                allMethods->numericGetElt = numericMethods->numericGetElt =
                    numericMethods->getExtElt = (getExtElt_t) addr;
                break;
            case INTSXP:
                allMethods->integerGetElt = integerMethods->integerGetElt =
                    integerMethods->getExtElt = (getExtElt_t) addr;
                break;
            case LGLSXP:
                allMethods->logicalGetElt = logicalMethods->logicalGetElt =
                    logicalMethods->getExtElt = logicalMethods->integerGetElt =
                    (getExtElt_t) addr;
                break;
            case CPLXSXP:
                allMethods->complexGetElt = complexMethods->complexGetElt =
                    complexMethods->getExtElt = (getExtElt_t) addr;
                break;
            case STRSXP:
                allMethods->characterGetElt =
                    characterMethods->characterGetElt =
                    characterMethods->getExtElt = (getExtElt_t) addr;
                break;
            case VECSXP:
                allMethods->anyGetElt =
                    listMethods->anyGetElt =
                    listMethods->getExtElt = (getExtElt_t) addr;
                break;
            default:
                break;
            }
            break;
        case SETELT_METHOD:
            switch (mode) {
            case REALSXP:
                allMethods->numericSetElt = numericMethods->numericSetElt =
                    numericMethods->setExtElt = (setExtElt_t) addr;
                break;
            case INTSXP:
                allMethods->integerSetElt = integerMethods->integerSetElt =
                    integerMethods->setExtElt = integerMethods->logicalSetElt =
                    (setExtElt_t) addr;
                break;
            case LGLSXP:
                allMethods->logicalSetElt = logicalMethods->logicalSetElt =
                    logicalMethods->setExtElt = (setExtElt_t) addr;
                break;
            case CPLXSXP:
                allMethods->complexSetElt = complexMethods->complexSetElt =
                    complexMethods->setExtElt = (setExtElt_t) addr;
                break;
            case STRSXP:
                allMethods->characterSetElt =
                    characterMethods->characterSetElt =
                    characterMethods->setExtElt = (setExtElt_t) addr;
                break;
            case VECSXP:
                allMethods->anySetElt =
                    listMethods->anySetElt =
                    listMethods->setExtElt = (setExtElt_t) addr;
                break;
            default:
                break;
            }
            break;
        case VECSUBSET_METHOD:
            switch (mode) {
            case REALSXP:
                allMethods->numericSubset = numericMethods->numericSubset =
                    numericMethods->vectorSubset = (extSubset_t) addr;
                break;
            case INTSXP:
                allMethods->integerSubset = integerMethods->integerSubset =
                    integerMethods->vectorSubset = (extSubset_t) addr;
                break;
            case LGLSXP:
                allMethods->logicalSubset = logicalMethods->logicalSubset =
                    logicalMethods->vectorSubset = (extSubset_t) addr;
                break;
            case CPLXSXP:
                allMethods->complexSubset = complexMethods->complexSubset =
                    complexMethods->vectorSubset = (extSubset_t) addr;
                break;
            case STRSXP:
                allMethods->characterSubset =
                    characterMethods->characterSubset =
                    characterMethods->vectorSubset = (extSubset_t) addr;
                break;
            case VECSXP:
                allMethods->anySubset =
                    listMethods->anySubset =
                    listMethods->vectorSubset = (extSubset_t) addr;
                break;
            default:
                break;
            }
            break;
        case VECSUBASSIGN_METHOD:
            switch (mode) {
            case REALSXP:
                allMethods->numericSubassign =
                    numericMethods->numericSubassign = (extSubassign_t) addr;
                break;
            case INTSXP:
                allMethods->integerSubassign =
                    integerMethods->integerSubassign = (extSubassign_t) addr;
                break;
            case LGLSXP:
                allMethods->logicalSubassign =
                    logicalMethods->logicalSubassign = (extSubassign_t) addr;
                break;
            case CPLXSXP:
                allMethods->complexSubassign =
                    complexMethods->complexSubassign = (extSubassign_t) addr;
                break;
            case STRSXP:
                allMethods->characterSubassign =
                    characterMethods->characterSubassign = (extSubassign_t) addr;
                break;
            case VECSXP:
                allMethods->anySubassign =
                    listMethods->anySubassign = (extSubassign_t) addr;
                break;
            default:
                break;
            }
            break;
        case MATGETELT_METHOD:
            switch (mode) {
            case REALSXP:
                allMethods->numericGetMatrixElt =
                    numericMethods->numericGetMatrixElt =
                    numericMethods->getMatrixElt =
                    (getExtMatrixElt_t) addr;
                break;
            case INTSXP:
                allMethods->integerGetMatrixElt =
                    integerMethods->integerGetMatrixElt =
                    integerMethods->getMatrixElt =
                    (getExtMatrixElt_t) addr;
                break;
            case LGLSXP:
                allMethods->logicalGetMatrixElt =
                    logicalMethods->logicalGetMatrixElt =
                    logicalMethods->getMatrixElt =
                    (getExtMatrixElt_t) addr;
                break;
            case CPLXSXP:
                allMethods->complexGetMatrixElt =
                    complexMethods->complexGetMatrixElt =
                    complexMethods->getMatrixElt =
                    (getExtMatrixElt_t) addr;
                break;
            case STRSXP:
                allMethods->characterGetMatrixElt =
                    characterMethods->characterGetMatrixElt =
                    characterMethods->getMatrixElt =
                    (getExtMatrixElt_t) addr;
                break;
            case VECSXP:
                allMethods->anyGetMatrixElt =
                    listMethods->anyGetMatrixElt =
                    listMethods->getMatrixElt =
                    (getExtMatrixElt_t) addr;
                break;
            default:
                break;
            }
            break;
        case MATSETELT_METHOD:
            switch (mode) {
            case REALSXP:
                allMethods->numericSetMatrixElt =
                    numericMethods->numericSetMatrixElt =
                    numericMethods->setMatrixElt =
                    (setExtMatrixElt_t) addr;
                break;
            case INTSXP:
                allMethods->integerSetMatrixElt =
                    integerMethods->integerSetMatrixElt =
                    integerMethods->setMatrixElt =
                    (setExtMatrixElt_t) addr;
                break;
            case LGLSXP:
                allMethods->logicalSetMatrixElt =
                    logicalMethods->logicalSetMatrixElt =
                    logicalMethods->setMatrixElt =
                    (setExtMatrixElt_t) addr;
                break;
            case CPLXSXP:
                allMethods->complexSetMatrixElt =
                    complexMethods->complexSetMatrixElt =
                    complexMethods->setMatrixElt =
                    (setExtMatrixElt_t) addr;
                break;
            case STRSXP:
                allMethods->characterSetMatrixElt =
                    characterMethods->characterSetMatrixElt =
                    characterMethods->setMatrixElt =
                    (setExtMatrixElt_t) addr;
                break;
            case VECSXP:
                allMethods->anySetMatrixElt =
                    listMethods->anySetMatrixElt =
                    listMethods->setMatrixElt =
                    (setExtMatrixElt_t) addr;
                break;
            default:
                break;
            }
            break;
        case MATSUBSET_METHOD:
            switch (mode) {
            case REALSXP:
                allMethods->numericMatrixSubset =
                    numericMethods->numericMatrixSubset =
                    numericMethods->matrixSubset =
                    (extMatrixSubset_t) addr;
                break;
            case INTSXP:
                allMethods->integerMatrixSubset =
                    integerMethods->integerMatrixSubset =
                    integerMethods->matrixSubset =
                    (extMatrixSubset_t) addr;
                break;
            case LGLSXP:
                allMethods->logicalMatrixSubset =
                    logicalMethods->logicalMatrixSubset =
                    logicalMethods->matrixSubset =
                    (extMatrixSubset_t) addr;
                break;
            case CPLXSXP:
                allMethods->complexMatrixSubset =
                    complexMethods->complexMatrixSubset =
                    complexMethods->matrixSubset =
                    (extMatrixSubset_t) addr;
                break;
            case STRSXP:
                allMethods->characterMatrixSubset =
                    characterMethods->characterMatrixSubset =
                    characterMethods->matrixSubset =
                    (extMatrixSubset_t) addr;
                break;
            case VECSXP:
                allMethods->anyMatrixSubset =
                    listMethods->anyMatrixSubset =
                    listMethods->matrixSubset =
                    (extMatrixSubset_t) addr;
                break;
            default:
                break;
            }
            break;
        case MATSUBASSIGN_METHOD:
            switch (mode) {
            case REALSXP:
                allMethods->numericMatrixSubassign =
                    numericMethods->numericMatrixSubassign =
                    (extMatrixSubassign_t) addr;
                break;
            case INTSXP:
                allMethods->integerMatrixSubassign =
                    integerMethods->integerMatrixSubassign =
                    (extMatrixSubassign_t) addr;
                break;
            case LGLSXP:
                allMethods->logicalMatrixSubassign =
                    logicalMethods->logicalMatrixSubassign =
                    (extMatrixSubassign_t) addr;
                break;
            case CPLXSXP:
                allMethods->complexMatrixSubassign =
                    complexMethods->complexMatrixSubassign =
                    (extMatrixSubassign_t) addr;
                break;
            case STRSXP:
                allMethods->characterMatrixSubassign =
                    characterMethods->characterMatrixSubassign =
                    (extMatrixSubassign_t) addr;
                break;
            case VECSXP:
                allMethods->anyMatrixSubassign =
                    listMethods->anyMatrixSubassign =
                    (extMatrixSubassign_t) addr;
                break;
            default:
                break;
            }
            break;
        default:
            warning("unrecognized native method %s", nm);
        }
    }
    return thisMethods;
}

/* methods for simpleStorage class */

static primitiveElt_t*
simpleLogicalGetElt(const SEXPREC* storage, const storageMethods_t* methods,
                    int i, primitiveElt_t* val, int* warn)
{
    val->logical = ((int *) R_ExternalPtrAddr((SEXP) storage))[i];
    return val;
}

static const primitiveElt_t*
simpleLogicalSetElt(SEXP storage, const storageMethods_t* methods,
                    int i, const primitiveElt_t* val, int* warn)
{
    int* tmp = (int*) R_ExternalPtrAddr((SEXP) storage);
    tmp[i] = val->logical;
    return val;
}

static primitiveElt_t*
simpleIntegerGetElt(const SEXPREC* storage, const storageMethods_t* methods,
                    int i, primitiveElt_t* val, int* warn)
{
    int* tmp = (int*) R_ExternalPtrAddr((SEXP) storage);
    val->integer = tmp[i];
    return val;
}

static const primitiveElt_t*
simpleIntegerSetElt(SEXP storage, const storageMethods_t* methods,
                    int i, const primitiveElt_t* val, int* warn)
{
    int* tmp = (int*) R_ExternalPtrAddr((SEXP) storage);
    tmp[i] = val->integer;
    return val;
}

static primitiveElt_t*
simpleNumericGetElt(const SEXPREC* storage, const storageMethods_t* methods,
                    int i, primitiveElt_t* val, int* warn)
{
    double* tmp = (double*) R_ExternalPtrAddr((SEXP) storage);
    val->numeric = tmp[i];
    return val;
}

static const primitiveElt_t*
simpleNumericSetElt(SEXP storage, const storageMethods_t* methods,
                    int i, const primitiveElt_t* val, int* warn)
{
    double* tmp = (double*) R_ExternalPtrAddr((SEXP) storage);
    tmp[i] = val->numeric;
    return val;
}

static primitiveElt_t*
simpleComplexGetElt(const SEXPREC* storage, const storageMethods_t* methods,
                    int i, primitiveElt_t* val, int* warn)
{
    Rcomplex* tmp = (Rcomplex*) R_ExternalPtrAddr((SEXP) storage);
    val->cmplx.r = tmp[i].r;
    val->cmplx.i = tmp[i].i;
    return val;
}

static const primitiveElt_t*
simpleComplexSetElt(SEXP storage, const storageMethods_t* methods,
                    int i, const primitiveElt_t* val, int* warn)
{
    Rcomplex* tmp = (Rcomplex*) R_ExternalPtrAddr(storage);
    tmp[i].r = val->cmplx.r;
    tmp[i].i = val->cmplx.i;
    return val;
}

static primitiveElt_t*
simpleImaginaryGetElt(const SEXPREC* storage, const storageMethods_t* methods,
                      int i, primitiveElt_t* val, int* warn)
{
    Rcomplex* tmp = (Rcomplex*) R_ExternalPtrAddr((SEXP) storage);
    val->numeric = tmp[i].i;
    return val;
}

static const primitiveElt_t*
simpleImaginarySetElt(SEXP storage, const storageMethods_t* methods,
                      int i, const primitiveElt_t* val, int* warn)
{
    Rcomplex* tmp = (Rcomplex*) R_ExternalPtrAddr(storage);
    tmp[i].i = val->numeric;
    return val;
}

static primitiveElt_t*
simpleCharacterGetElt(const SEXPREC* storage, const storageMethods_t* methods,
                      int i, primitiveElt_t* val, int* warn)
{
    SEXP vec = R_ExternalPtrProtected((SEXP) storage);
    val->character = STRING_ELT(vec, i);
    return val;
}

static const primitiveElt_t*
simpleCharacterSetElt(SEXP storage, const storageMethods_t* methods,
                      int i, const primitiveElt_t* val, int* warn)
{
    SEXP vec = R_ExternalPtrProtected((SEXP) storage);
    SET_STRING_ELT(vec, i, val->any);
    return val;
}

static primitiveElt_t*
simpleAnyGetElt(const SEXPREC* storage, const storageMethods_t* methods,
                int i, primitiveElt_t* val, int* warn)
{
    SEXP vec = R_ExternalPtrProtected((SEXP) storage);
    val->any = VECTOR_ELT(vec, i);
    return val;
}

static const primitiveElt_t*
simpleAnySetElt(SEXP storage, const storageMethods_t* methods,
                int i, const primitiveElt_t* val, int* warn)
{
    SEXP vec = R_ExternalPtrProtected((SEXP) storage);
    SET_VECTOR_ELT(vec, i, val->any);
    return val;
}

static int
simpleSize(const SEXPREC* x)
{
    return asInteger(GET_SLOT((SEXP)x, install("length")));
}

static SEXP
simpleResize(SEXP x,  int n)
{
    gcExternalPtr_resizeInternal(GET_STORAGE_EXT_PTR(x), n, 1);
    return x;
}


#include <R.h>
#include <R_ext/Rdynload.h>

static const R_CallMethodDef R_CallDef  [] = {
    CALL_DEF(createNativeStorageMethods, 3),
    CALL_DEF(storageMethodsStructSize, 0),
    CALL_DEF(initNativeStorageMethods, 1),
    {NULL, NULL, 0},
};

static const R_CMethodDef R_CDef[]  = {
  {"logicalGetElt", (DL_FUNC) &simpleLogicalGetElt, 5, NULL, NULL},
  {"integerGetElt", (DL_FUNC) &simpleIntegerGetElt, 5, NULL, NULL},
  {"numericGetElt", (DL_FUNC) &simpleNumericGetElt, 5, NULL, NULL},
  {"complexGetElt", (DL_FUNC) &simpleComplexGetElt, 5, NULL, NULL},
  {"imaginaryGetElt", (DL_FUNC) &simpleImaginaryGetElt, 5, NULL, NULL},
  {"characterGetElt", (DL_FUNC) &simpleCharacterGetElt, 5, NULL, NULL},
  {"anyGetElt", (DL_FUNC) &simpleAnyGetElt, 5, NULL, NULL},
  {"logicalSetElt", (DL_FUNC) &simpleLogicalSetElt, 5, NULL, NULL},
  {"integerSetElt", (DL_FUNC) &simpleIntegerSetElt, 5, NULL, NULL},
  {"numericSetElt", (DL_FUNC) &simpleNumericSetElt, 5, NULL, NULL},
  {"complexSetElt", (DL_FUNC) &simpleComplexSetElt, 5, NULL, NULL},
  {"imaginarySetElt", (DL_FUNC) &simpleImaginarySetElt, 5, NULL, NULL},
  {"characterSetElt", (DL_FUNC) &simpleCharacterSetElt, 5, NULL, NULL},
  {"anySetElt", (DL_FUNC) &simpleAnySetElt, 5, NULL, NULL},
  {"size", (DL_FUNC) &simpleSize, 2, NULL, NULL},
  {"resize", (DL_FUNC) &simpleResize, 3, NULL, NULL},
  {NULL, NULL, 0, NULL, NULL}
};


const R_CMethodDef*
storage_C_methods(int* nptr)
{
    int n;
    R_CMethodDef* ans;
    for (n = 0; R_CDef[n].name != NULL; n++) {;}
    *nptr = n;
    ans = Calloc(++n, R_CMethodDef);
    memcpy(ans, R_CDef, n*sizeof(R_CMethodDef));
    return ans;
}


const R_CallMethodDef*
storage_Call_methods(int* nptr)
{
    *nptr = LENGTH_OF_METHODS_ARRAY(R_CallDef);
    return &R_CallDef[0];
}
