#include "subscript.h"
#include "storageMethods.h"

#ifdef VectorSubscript
#undef VectorSubscript
#endif
#ifdef ArraySubscript
#undef ArraySubscript
#endif
#define VectorSubscript localVectorSubscript
#define ArraySubscript localArraySubscript

/* Start of cut and paste from R/src/main/subscript.c */

static SEXP nullSubscript(int n)
{
    int i;
    SEXP indx;
    indx = allocVector(INTSXP, n);
    for (i = 0; i < n; i++)
	INTEGER(indx)[i] = i + 1;
    return indx;
}

static SEXP logicalSubscript(SEXP s, int ns, int nx, int *stretch,
                             int assign)
{
    int canstretch, count, i, nmax;
    SEXP indx;
    canstretch = *stretch;
    if (!canstretch && ns > nx)
        error("(subscript) logical subscript too long");
    nmax = (ns > nx) ? ns : nx;
    *stretch = (ns > nx) ? ns : 0;
    if (ns == 0)
        return(allocVector(INTSXP, 0));
    count = 0;
    if (assign) {
        for (i = 0; i < nmax; i++)
            if (LOGICAL(s)[i%ns] == TRUE)
                count++;
    } else {
        for (i = 0; i < nmax; i++)
            if (LOGICAL(s)[i%ns] != FALSE)
                count++;
    }
    indx = allocVector(INTSXP, count);
    count = 0;
    if (assign) {
        for (i = 0; i < nmax; i++)
            if (LOGICAL(s)[i%ns] == TRUE)
                INTEGER(indx)[count++] = i + 1;
    } else {
        for (i = 0; i < nmax; i++) {
            int tmp = LOGICAL(s)[i%ns];
            if (tmp == NA_LOGICAL)
                INTEGER(indx)[count++] = NA_INTEGER;
            else if (tmp == TRUE) INTEGER(indx)[count++] = i + 1;
        }
    }
    return indx;
}

static SEXP negativeSubscript(SEXP s, int ns, int nx)
{
    int count, i;
    SEXP indx = PROTECT(allocVector(LGLSXP, nx));

    for (i = 0; i < nx; i++)
        LOGICAL(indx)[i] = TRUE;
    count = nx;
    for (i = 0; i < ns; i++)
        if (INTEGER(s)[i] != 0) {
            LOGICAL(indx)[-INTEGER(s)[i] - 1] = FALSE;
            count--;
        }
    s = allocVector(INTSXP, count);
    count = 0;
    for (i = 0; i < nx; i++)
        if (LOGICAL(indx)[i])
            INTEGER(s)[count++] = i + 1;

    UNPROTECT(1);
    return s;
}

static SEXP positiveSubscript(SEXP s, int ns, int nx, int remove_na)
{
    SEXP indx;
    int i, zct = 0;
    if (remove_na) {
        for (i = 0; i < ns; i++) {
            int tmp = INTEGER(s)[i];
            if (tmp == 0 || tmp == NA_INTEGER)
                zct++;
        }
        if (zct) {
            indx = allocVector(INTSXP, (ns - zct));
            for (i = 0, zct = 0; i < ns; i++) {
                int tmp = INTEGER(s)[i];
                if (tmp != 0 && tmp != NA_INTEGER)
                    INTEGER(indx)[zct++] = tmp;
            }
            return indx;
        } else return s;
    } else {
        for (i = 0; i < ns; i++) {
            if (INTEGER(s)[i] == 0)
                zct++;
        }
        if (zct) {
            indx = allocVector(INTSXP, (ns - zct));
            for (i = 0, zct = 0; i < ns; i++)
                if (INTEGER(s)[i] != 0)
                    INTEGER(indx)[zct++] = INTEGER(s)[i];
            return indx;
        } else return s;
    }
}

static SEXP integerSubscript(SEXP s, int ns, int nx, int *stretch,
                             int assign)
{
    int i, ii, min, max, canstretch;
    int have_na = 0;
    canstretch = *stretch;
    *stretch = 0;
    min = 1;
    max = 0;
    for (i = 0; i < ns; i++) {
        ii = INTEGER(s)[i];
        if (ii == NA_INTEGER) {
            have_na = 1;
        } else {
            if (ii < min)
                min = ii;
            if (ii > max)
                max = ii;
        }
    }
    if (min < -nx)
        error("subscript out of bounds");
    if (max > nx) {
        if(canstretch) *stretch = max;
        else error("subscript out of bounds");
    }
    if (min < 0) {
        if (max != 0 || have_na)
            error("only 0's may mix with negative subscripts");
        return negativeSubscript(s, ns, nx);
    }
    assign = assign && have_na;
    if (assign || min == 0)
        return positiveSubscript(s, ns, nx, assign);
    else return s;
}

typedef SEXP (*StringEltGetter)(SEXP x, int i); 

static SEXP stringSubscript(SEXP s, int ns, int nx, SEXP names,
                            StringEltGetter getElt,
                            int *stretch, int assign)
{
    SEXP indx, indexnames;
    int i, j, nnames, sub, extra;
    int canstretch = *stretch;
    PROTECT(s);
    PROTECT(names);
    PROTECT(indx = allocVector(INTSXP, ns));
    PROTECT(indexnames = allocVector(STRSXP, ns));
    nnames = nx;
    extra = nnames;

    /* Process each of the subscripts */
    /* First we compare with the names on the vector */
    /* and then (if there is no match) with each of */
    /* the previous subscripts. */

    for (i = 0; i < ns; i++) {
	sub = 0;
	if (names != R_NilValue) {
	    for (j = 0; j < nnames; j++)
		if (NonNullStringMatch(STRING_ELT(s, i), getElt(names, j))) {
		    sub = j + 1;
		    SET_STRING_ELT(indexnames, i, R_NilValue);
		    break;
		}
	}
	if (sub == 0) {
	    for (j = 0 ; j < i ; j++)
		if (NonNullStringMatch(STRING_ELT(s, i), STRING_ELT(s, j))) {
		    sub = INTEGER(indx)[j];
/*		    SET_STRING_ELT(indexnames, i, STRING_ELT(indexnames, sub - 1));*/
		    SET_STRING_ELT(indexnames, i, STRING_ELT(s, j));
		    break;
		}
	}
	if (sub == 0) {
	    if (!canstretch)
		error("subscript out of bounds");
	    extra += 1;
	    sub = extra;
	    SET_STRING_ELT(indexnames, i, STRING_ELT(s, i));
	}
	INTEGER(indx)[i] = sub;
    }
    /* Ghastly hack!  We attach the new names to the attribute */
    /* slot on the returned subscript vector. */
    if (extra != nnames) {
	SET_ATTRIB(indx, indexnames);
    }
    if (canstretch)
	*stretch = extra;
    UNPROTECT(4);
    return indx;
}

/* Array Subscripts.  
    dim is the dimension (0 to k-1)
    s is the subscript list, 
    dims is the dimensions of x
    dng is a function (usually getAttrib) that obtains the dimnames
    x is the array to be subscripted. 
*/

typedef SEXP (*AttrGetter)(SEXP x, SEXP data);

SEXP ArraySubscript(int dim, SEXP s, SEXP dims, int assign,
                    AttrGetter dng, StringEltGetter getElt, SEXP x)
{
    int nd, ns, stretch = 0;
    SEXP dnames, tmp;
    ns = length(s);
    nd = INTEGER(dims)[dim];

    switch (TYPEOF(s)) {
    case NILSXP:
	return allocVector(INTSXP, 0);
    case LGLSXP:
	return logicalSubscript(s, ns, nd, &stretch, assign);
    case INTSXP:
	return integerSubscript(s, ns, nd, &stretch, assign);
    case REALSXP:
    	PROTECT(tmp = coerceVector(s, INTSXP));
	tmp = integerSubscript(tmp, ns, nd, &stretch, assign);
    	UNPROTECT(1);
	return tmp;
    case STRSXP:
	dnames = dng(x, R_DimNamesSymbol);
	if (dnames == R_NilValue)
	    error("no dimnames attribute for array");
	dnames = VECTOR_ELT(dnames, dim);
	return stringSubscript(s, ns, nd, dnames, getElt,
                           &stretch, assign);
    case SYMSXP:
	if (s == R_MissingArg)
	    return nullSubscript(nd);
    default:
	error("invalid subscript");
    }
    return R_NilValue;
}

/* nx is the length of the object being subscripted,
   s is the R subscript value,
   dng gets a given attrib for x, which is the object we are
   subsetting,
*/
 
SEXP VectorSubscript(int nx, SEXP s, int *stretch, int assign,
                     AttrGetter dng, StringEltGetter getElt, SEXP x)
{
    int ns;
    SEXP ans=R_NilValue, tmp;

    ns = length(s);
    /* special case for simple indices -- does not duplicate */
    if (ns == 1 && TYPEOF(s) == INTSXP && ATTRIB(s) == R_NilValue) {
	int i = INTEGER(s)[0];
	if (0 < i && i <= nx) {
	    *stretch = 0;
	    return s;
	}
    }
    PROTECT(s=duplicate(s));
    SET_ATTRIB(s, R_NilValue);
    switch (TYPEOF(s)) {
    case NILSXP:
	*stretch = 0;
	ans = allocVector(INTSXP, 0);
	break;
    case LGLSXP:
	/* *stretch = 0; */
	ans = logicalSubscript(s, ns, nx, stretch, assign);
	break;
    case INTSXP:
	    ans = integerSubscript(s, ns, nx, stretch, assign);
	    break;
    case REALSXP:
	PROTECT(tmp = coerceVector(s, INTSXP));
	ans = integerSubscript(tmp, ns, nx, stretch, assign);
	UNPROTECT(1);
	break;
    case STRSXP:
    {
	SEXP names = dng(x, R_NamesSymbol);
	/* *stretch = 0; */
	ans = stringSubscript(s, ns, nx, names, getElt,
                          stretch, assign);
    }
    break;
    case SYMSXP:
	*stretch = 0;
	if (s == R_MissingArg) {
	    ans = nullSubscript(nx);
	    break;
	}
    default:
	error("invalid subscript type");
    }
    UNPROTECT(1);
    return ans;
}

/* end of cut and paste from R/src/main/subscript.c */

static SEXP myArraySubscript(int dim, SEXP s, SEXP dims, int assign,
                             AttrGetter dng, StringEltGetter getElt,
                             SEXP x)
{
    int n = INTEGER(dims)[dim];
    const SEXPREC* stype;
    if (s == R_MissingArg)
        return allIndex;
    stype = getIndexType(s);
    if (s != stype || s == allIndex || s == noneIndex) {
        return s;
    }
    if (isSimpleSubscript(s, n, assign?REMOVE_NA:KEEP_NA, 0)) {
        s = simplifyProcessedSubscript(s, NO_NA, 0, 1, n);
    } else {
        s = ArraySubscript(dim, s, dims, assign, dng,
                           getElt, x);
        PROTECT(s);
        SET_ATTRIB(s, R_NilValue);
        s = simplifyProcessedSubscript(s, assign?NO_NA:KEEP_NA, 0, 1, n);
        UNPROTECT(1);
    }
    return s;
}

static SEXP myVectorSubscript(int nx, SEXP s, int *stretch, int assign,
                              AttrGetter dng, StringEltGetter getElt,
                              SEXP x, SEXP* newnames)
{
    int canstretch = *stretch;
    const SEXPREC* stype;
    *newnames = R_NilValue;
    *stretch = 0;
    if (s == R_MissingArg) {
        return allIndex;
    }
    stype = getIndexType(s);
    if (s != stype || s == allIndex || s == noneIndex) {
        return s;
    }
    if (isSimpleSubscript(s, nx, assign?REMOVE_NA:KEEP_NA, canstretch)) {
         return simplifyProcessedSubscript(s, NO_NA, 0, 1, nx);
    }
    s = VectorSubscript(nx, s, &canstretch, assign, dng, getElt, x);
    PROTECT(s);
    *newnames = ATTRIB(s);
    SET_ATTRIB(s, R_NilValue);
    s = simplifyProcessedSubscript(s,
                                   assign?NO_NA:KEEP_NA,
                                   canstretch, 1, nx);
    UNPROTECT(1);
    *stretch = canstretch;
    return s;
}

static SEXP identity(SEXP x, SEXP data) { return x; }

static SEXP getExtCharElt(SEXP x, int i)
{
    primitiveElt_t elt;
    storageMethods_t* methods = getStorageMethods(x);
    SEXP storage = GET_STORAGE_EXT_PTR_UNCHECKED(x);
    int warn = 0;

    methods->characterGetElt(storage, methods, i+1, &elt, &warn);
    CoercionWarning(warn);
    return elt.character;
}

SEXP
SubscriptList(SEXP subs, SEXP len, SEXP dims, SEXP names,
              SEXP dimnames, SEXP assign)
{
    int assignop = asLogical(assign);
    int n;
    int ndim = LENGTH(subs);
    SEXP ans;
    StringEltGetter getElt;
    if (ndim == 0) {
        ans = allocVector(VECSXP, 1);
        PROTECT(ans);
        SET_VECTOR_ELT(ans, 0, allIndex);
        setAttrib(ans, install("newlength"), len);
    } else if (ndim == 1) {
        SEXP newnames;
        int stretch = 1;
        if (isExternalVectorWithStorage(names)) {
            names = GET_SLOT(names, install("storage"));
            /* force restoring of external pointer */
            GET_STORAGE_EXT_PTR(names);
            getElt = getExtCharElt;
        } else {
            if (names != R_NilValue)
                names = coerceVector(names, STRSXP);
            getElt = STRING_ELT;
        }
        PROTECT(names);
        ans = allocVector(VECSXP, 1);
        PROTECT(ans);
        SET_VECTOR_ELT(ans, 0, myVectorSubscript(asInteger(len),
                                                 VECTOR_ELT(subs, 0),
                                                 &stretch, assignop,
                                                 identity,
                                                 getElt, names,
                                                 &newnames));
        if (stretch > 0) {
            setAttrib(ans, install("newlength"), ScalarInteger(stretch));
        } else {
            setAttrib(ans, install("newlength"), len);
        }
        if (newnames != R_NilValue) {
            setAttrib(ans, install("newnames"), newnames);
        }
        UNPROTECT(1);
    } else {
        int* idim;
        int i;

        dims = coerceVector(dims, INTSXP);
        PROTECT(dims);
        idim = INTEGER(dims);
        if (LENGTH(dims) != ndim)
            error("incorrect number of dimensions");

        if (dimnames != R_NilValue) {
            SEXP tmp = allocVector(VECSXP, ndim);
            PROTECT(tmp);
            for (i = 0; i < ndim; i++) {
                names = VECTOR_ELT(dimnames, i);
                if (!isExternalVectorWithStorage(names))
                    names = coerceVector(names, STRSXP);
                SET_VECTOR_ELT(tmp, i, names);
            }
            dimnames = tmp;
        } else PROTECT(dimnames);

        getElt = STRING_ELT; /* -Wall */

        ans = allocVector(VECSXP, ndim);
        PROTECT(ans);
        for (i = 0; i < ndim; i++) {
            n = idim[i];
            if (dimnames != R_NilValue) {
                names = VECTOR_ELT(dimnames, i);
                if (isExternalVectorWithStorage(names)) {
                    names = GET_SLOT(names, install("storage"));
                    /* force restoring of external pointer */
                    GET_STORAGE_EXT_PTR(names);
                    getElt = getExtCharElt;
                } else {
                    getElt = STRING_ELT;
                }
            }
            SET_VECTOR_ELT(ans, i, myArraySubscript(i,
                                                    VECTOR_ELT(subs, i),
                                                    dims, assignop,
                                                    identity,
                                                    getElt, dimnames));
        }
        UNPROTECT(2);
    }
    UNPROTECT(1);
    return ans;
}

#include <R.h>
#include <R_ext/Rdynload.h>

static const R_CallMethodDef R_CallDef  [] = {
    CALL_DEF(SubscriptList, 6),
   {NULL, NULL, 0},
};

const R_CallMethodDef*
subscript_Call_methods(int* nptr)
{
    *nptr = LENGTH_OF_METHODS_ARRAY(R_CallDef);
    return &R_CallDef[0];
}
